import os
import matplotlib
from matplotlib import pyplot as plt
from matplotlib.patches import Circle
from matplotlib.collections import PatchCollection
import matplotlib.colors as mcolors
import numpy as np
from vipy.util import islist, temppng
import sys
from vipy.globals import log
from vipy.version import Version
import numpy as np


FIGHANDLE = {}
matplotlib.rcParams['toolbar'] = 'None'
matplotlib_version_at_least_3p3 = Version.from_string(matplotlib.__version__) >= '3.3'


PRIMARY_COLORLIST = ['green','blue','red','cyan','orange','yellow','violet','white'] + ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan']
COLORLIST = PRIMARY_COLORLIST + [str(name) for (name, hex) in mcolors.cnames.items() if str(name) not in PRIMARY_COLORLIST]  # use primary colors first
COLORLIST_LUMINANCE = [(np.array(mcolors.to_rgb(name)) * np.array([0.2126, 0.7152, 0.0722])).sum() for name in COLORLIST]
DARK_COLORLIST = tuple(c for (c,l) in zip(COLORLIST, COLORLIST_LUMINANCE) if l>=0.5)  # colors that show well on dark backgrounds
LIGHT_COLORLIST = tuple(c for (c,l) in zip(COLORLIST, COLORLIST_LUMINANCE) if l<=0.5) # colors that show well on light backgrounds


# Optional latex strings in captions
try:
    import vipy.globals
    if vipy.globals.GLOBAL['LATEX'] is not None:
        from distutils.spawn import find_executable
        if not find_executable('latex'):
            raise
        matplotlib.rc('text', usetex=True)  # requires latex installed
except:
    pass  # ignored if latex is not installed or not wanted


def escape_to_exit(event):
    if event.key == 'escape' or event.key == 'q' or event.key == 'ctrl+c':
        import vipy.globals            
        vipy.globals._user_hit_escape(True)
    
def flush():
    if matplotlib.get_backend().lower() == 'macosx':
        plt.draw()  # YUCK: to flush buffer on video play, will be slow
    plt.pause(0.001)

    
def imflush():
    if matplotlib.get_backend().lower() == 'macosx':
        plt.draw()  # YUCK: to flush buffer on video play, will be slow
    plt.pause(0.001)

    # this is necessary for imshow only, maybe to trigger remove() of child?
    # However, this breaks jupyter notebook image show if we use gca()
    [a.annotate('', (0,0)) for i in plt.get_fignums() for a in plt.figure(i).axes]
    
    #plt.gca().annotate('', (0,0))  

    
def show(fignum):
    fig = plt.figure(fignum) 
    fig.canvas.draw()    
    plt.ion()
    plt.show()

    
def noshow(fignum):
    plt.ioff()


def savefig(filename=None, fignum=None, pad_inches=0, dpi=None, bbox_inches='tight', format=None):
    if fignum is not None:
        plt.figure(fignum)
    if filename is None:
        filename = temppng()
    plt.savefig(filename, pad_inches=pad_inches, dpi=dpi, bbox_inches=bbox_inches, format=format)
    return filename


def figure(fignum=None):
    if fignum is not None:
        plt.figure(fignum)
    else:
        plt.figure()
    return plt


def close(fignum):
    global FIGHANDLE
    if fignum in FIGHANDLE:
        plt.close(fignum)
        FIGHANDLE.pop(fignum, None)
        return None
    else:
        return None


def closeall():
    global FIGHANDLE
    FIGHANDLE = {}
    return plt.close('all')


def _imshow_tight(img, fignum=None, keypress=True):
    """Helper function to show an image in a figure window"""
    dpi = 100.0
    fig = plt.figure(fignum, dpi=dpi, figsize=(img.shape[1] / dpi, img.shape[0] / dpi))
    plt.clf()

    # Tight axes
    ax = plt.Axes(fig, [0., 0., 1.0, 1.0], frameon=False)
    fig.add_axes(ax)
    for a in plt.gcf().axes:
        a.get_xaxis().set_visible(False)
        a.get_yaxis().set_visible(False)
    imh = plt.imshow(img, animated=True, interpolation='nearest', aspect='equal')

    if keypress:
        fig.canvas.mpl_connect('key_press_event', escape_to_exit)
        import vipy.globals    
        vipy.globals._user_hit_escape(False)

    return (fig.number, imh)


def imshow(img, fignum=None):
    """Show an image in a figure window (optionally visible), reuse previous figure if it is the same shape"""
    global FIGHANDLE

    if fignum in plt.get_fignums() and fignum in FIGHANDLE and FIGHANDLE[fignum].get_size() == img.shape[0:2]:
        # Do not delete and recreate the figure, just change the pixels 
        FIGHANDLE[fignum].set_data(img)

        # Delete all polygon and text overlays from previous drawing so that they can be overwritten on current frame
        for c in plt.gca().get_children():
            desc = c.__repr__()
            if 'Text' in desc or 'Polygon' in desc or 'Circle' in desc or 'Line' in desc or 'Patch' in desc or 'PathCollection' in desc:
                try:
                    c.remove()
                except:
                    # MacOSX fails with the error: NotImplementedError: cannot remove artist
                    # fallback on closing figure completely
                    (fignum, imh) = _imshow_tight(img, fignum=fignum) 
                    FIGHANDLE[fignum] = imh
                    return fignum

    else:
        # Jupyter notebook does not respect fignum.  It is always one when inspecting plt.get_fignums()
        if fignum in plt.get_fignums() and fignum in FIGHANDLE:
            close(fignum)
            #pass  # don't close unless user requests, this is faster, but window size does not change with image resolution
        (fignum, imh) = _imshow_tight(img, fignum=fignum)
        FIGHANDLE[fignum] = imh
    return fignum


def text(caption, xmin, ymin, fignum=None, textcolor='black', textfacecolor='white', textfacealpha=1.0, fontsize=10, linewidth=3, facecolor='white', facealpha=0.5, alpha=1.0, pad=0.5, captionoffset=0):
    plt.figure(fignum) if fignum is not None else plt.gcf()
    lw = linewidth  # pull in the boxes by linewidth so that they do not overhang the figure

    newlines = caption.count('\n')
    
    # MatplotlibDeprecationWarning: The 's' parameter of annotate() has been renamed 'text' since Matplotlib 3.3            
    handle = plt.annotate(**{'text' if matplotlib_version_at_least_3p3 else 's': caption},
                          alpha=alpha, xy=(xmin,ymin), xytext=(xmin, ymin+captionoffset), xycoords='data',
                          color=textcolor, bbox=None if textfacecolor is None else dict(facecolor=textfacecolor, edgecolor=None, alpha=textfacealpha, boxstyle='square', pad=pad),
                          fontsize=fontsize, clip_on=True)

    return fignum

def boundingbox(img, xmin, ymin, xmax, ymax, bboxcaption=None, fignum=None, bboxcolor='green', facecolor='white', facealpha=0.5, textcolor='black', textfacecolor='white', textfacealpha=1.0, fontsize=10, captionoffset=(0,0), linewidth=3):
    """Draw a captioned bounding box on a previously shown image"""
    plt.figure(fignum)
    ax = plt.gca()
    lw = linewidth  # pull in the boxes by linewidth so that they do not overhang the figure
    (H,W) = (img.shape[0], img.shape[1])
    
    ymin_frac = 1.0 - (min(ymax, H - lw/2) / float(H))
    ymax_frac = 1.0 - (max(ymin, lw/2) / float(H))
    
    y0, y1 = ax.get_ylim()                 # data‐coordinates of bottom and top of the axes
    rect = Rectangle((xmin, ymin), xmax - xmin,  (ymax_frac - ymin_frac) * abs(y1 - y0), 
                 linewidth=lw,
                 edgecolor=(*mcolors.to_rgb(bboxcolor),0.4),
                 facecolor=(*mcolors.to_rgb(facecolor),facealpha),
                 #alpha=facealpha,
                 transform=ax.transData, capstyle='round', joinstyle='bevel', clip_on=True)
    ax.add_patch(rect)

    # Text string
    if bboxcaption is not None:
        # a) Compute x‐fraction: (xmin − x0)/(x1 − x0)
        x0, x1 = ax.get_xlim()
        xmin_frac = ((xmin - x0) / (x1 - x0)) #+ (captionoffset[0]/H)
        
        # b) y1_frac is already the top of the box in axes‐fraction; add vertical offset
        ymin_frac_text = ymin_frac #+ (captionoffset[1]/W)
        ymax_frac = 1.0 - ((max(ymin, lw/2)) / float(H)) 
        
        newlines = bboxcaption.count('\n')

        captionoffset = captionoffset if ymin > 20 else (captionoffset[0], captionoffset[1]+(20*(1)))  # move down a bit if near top of image, shift once per newline in caption

        # MatplotlibDeprecationWarning: The 's' parameter of annotate() has been renamed 'text' since Matplotlib 3.3   
        plt.annotate(**{'text' if matplotlib_version_at_least_3p3 else 's': bboxcaption},
                     xy=(xmin_frac, ymax_frac),
                     xytext=(captionoffset[0], -captionoffset[1]),
                     textcoords='offset pixels',
                     xycoords='axes fraction',
                     color=textcolor,
                     fontsize=fontsize,
                     bbox=dict(
                         facecolor=textfacecolor,
                         edgecolor=textcolor,
                         alpha=textfacealpha,
                         boxstyle='square',
                     ),
                     horizontalalignment='left',
                     verticalalignment='top',
                     clip_on=True
                     )

    return fignum


def imdetection(img, detlist, fignum=None, bboxcolor='green', do_caption=True, facecolor='white', facealpha=0.5, textcolor='green', textfacecolor='white', textfacealpha=1.0, fontsize=10, captionoffset=(0,0)):
    """Show bounding boxes from a list of vipy.object.Detections on the same image, plotted in list order with optional captions """

    # Create image
    #fignum = imshow(img, fignum=fignum)  # this can fail on MacOSX, go the slow route instead
    fignum = _imshow_tight(img, fignum=fignum)     

    # A better way? https://matplotlib.org/api/_as_gen/matplotlib.animation.FuncAnimation.html
    
    # Valid detections
    for (k,det) in enumerate(detlist):
        if do_caption and det.category() is not None and len(det.category()) != 0 and (det.category()[0:2] != '__'):  # prepending category with '__' will disable caption
            bboxcaption = det.category()
        else:
            bboxcaption = None

        if islist(bboxcolor):
            bboxcolor_ = bboxcolor[k]
        else:
            bboxcolor_ = bboxcolor

        if islist(textcolor):
            textcolor_ = textcolor[k]
        else:
            textcolor_ = textcolor

        boundingbox(img, xmin=det.xmin(), ymin=det.ymin(), xmax=det.xmax(), ymax=det.ymax(), bboxcaption=bboxcaption,
                    fignum=fignum, bboxcolor=bboxcolor_, facecolor=facecolor, facealpha=facealpha, textcolor=textcolor_, textfacecolor=textfacecolor, fontsize=fontsize, captionoffset=captionoffset, textfacealpha=textfacealpha)

    return fignum


def imkeypoints(img, kplist, fignum=None, bordercolor='green', do_caption=True, facecolor='white', facealpha=0.5, textcolor='green', textfacecolor='white', textfacealpha=1.0, fontsize=10, captionoffset=(0,0)):
    """Show `vipy.object.Keypoint2d` on the same image, plotted in list order with optional captions """

    # Create image
    fignum = imshow(img, fignum=fignum) 

    # A better way? https://matplotlib.org/api/_as_gen/matplotlib.animation.FuncAnimation.html
    
    # Valid keypoints
    (x,y,size,color,caption,captioncolor) = ([],[],[],[],[],[])
    for (k,p) in enumerate(kplist):
        x.append(p.x)
        y.append(p.y)
        size.append(max(p.r, 1)**2)  # size is pts squared
        
        if do_caption and p.category() is not None and len(p.category()) != 0 and (p.category()[0:2] != '__'):  # prepending category with '__' will disable caption
            caption.append(p.category())
        else:
            caption.append(None)

        if islist(bordercolor):
            color.append(mcolors.to_hex(bordercolor[k]) + 'BB')  # with alpha
        else:
            color.append(mcolors.to_hex(bordercolor) + 'BB')  # with alpha
            
        if islist(textcolor):
            captioncolor.append(textcolor[k])
        else:
            captioncolor.append(textcolor)

    plt.figure(fignum)
    plt.scatter(x, y, c=color, s=size)
    return fignum


def imobjects(img, objlist, fignum=None, bordercolor='green', do_caption=True, facecolor='white', facealpha=0.5, textcolor='green', textfacecolor='white', textfacealpha=1.0, fontsize=10, captionoffset=(0,0), kp_alpha=0.8):
    """Show `vipy.object.Keypoint2d` on the same image, plotted in list order with optional captions """

    # Create image
    fignum = imshow(img, fignum=fignum) 
    plt.gca().set_autoscale_on(False)


    # A better way? https://matplotlib.org/api/_as_gen/matplotlib.animation.FuncAnimation.html

    # Valid detections
    detlist = [p for p in objlist if isinstance(p, vipy.object.Detection)]
    if len(detlist) > 0:
        for (k,det) in enumerate(detlist):
            if do_caption and det.category() is not None and len(det.category()) != 0 and (det.category()[0:2] != '__'):  # prepending category with '__' will disable caption
                bboxcaption = det.category()
            else:
                bboxcaption = None

            if islist(bordercolor):
                bboxcolor_ = bordercolor[k]
            else:
                bboxcolor_ = bordercolor

            if islist(textcolor):
                textcolor_ = textcolor[k]
            else:
                textcolor_ = textcolor

            boundingbox(img, xmin=det.xmin(), ymin=det.ymin(), xmax=det.xmax(), ymax=det.ymax(), bboxcaption=bboxcaption,
                        fignum=fignum, bboxcolor=bboxcolor_, facecolor=bboxcolor_, facealpha=facealpha, textcolor=textcolor_, textfacecolor=textfacecolor, fontsize=fontsize, captionoffset=captionoffset, textfacealpha=textfacealpha)
    
    # Valid keypoints
    kplist = [p for p in objlist if isinstance(p, vipy.object.Keypoint2d)]    
    plt.gca().set_autoscale_on(False)

    if len(kplist) > 0:
        (x,y,size,colors,r,patches) = ([],[],[],[],[],[])
        for (k,p) in enumerate(kplist):
            x.append(p.x)
            y.append(p.y)
            r.append(p.r)
            
            c = mcolors.to_hex(bordercolor[k] if islist(bordercolor) else bordercolor)
            colors.append(c + format(int(255*kp_alpha), '02X'))  # with alpha
            patches.append(Circle((p.x, p.y), p.r))

        plt.gca().add_collection(PatchCollection(patches, facecolors=colors, edgecolors='silver', clip_on=True, linewidths=0.5))  # scale radius with window size
    return fignum


def imframe(img, fr, color='b', markersize=10, label=None, figure=None):
    """Show a scatterplot of fr=[[x1,y1],[x2,y2]...] 2D points overlayed on an image, all the same color"""
    if figure is not None:
        fig = plt.figure(figure)
    else:
        fig = plt.figure()

    figure = plt.gcf().number

    plt.clf()

    ax = plt.Axes(fig, [0., 0., 1., 1.], frameon=False)
    fig.add_axes(ax)

    plt.axis('off')
    ax.set_axis_off()
    for a in plt.gcf().axes:
        a.get_xaxis().set_visible(False)
        a.get_yaxis().set_visible(False)

    plt.autoscale(tight=True)
    plt.plot(fr[:,0],fr[:,1],'%s.' % color, markersize=markersize, axes=ax)

    if label is not None:
        for ((x,y),lbl) in zip(fr, label):
            ax.text(x, y, lbl, color='white')

    return plt


def frame(fr, im=None, color='b.', markersize=10, figure=None, caption=None):
    """Show a scatterplot of fr=[[x1,y1],[x2,y2]...] 2D points, all the same color"""
    if figure is not None:
        plt.figure(figure)
    else:
        plt.figure()
        plt.clf()

    plt.axes([0,0,1,1])
    plt.plot(fr[:,0],fr[:,1],color)
    plt.axis('off')
    plt.draw()


def colorlist(theme=None):
    """Return a list of named colors that are higher contrast with a white background if light_mode, else named colors that are higher contrast with a dark background if dark_mode"""        
    return DARK_COLORLIST if theme=='dark' else (LIGHT_COLORLIST if theme=='light' else COLORLIST)


def edit():
    import matplotlib.pyplot as plt 
    from matplotlib.widgets import Slider, Button, RadioButtons   
    button = Button(resetax, 'Reset', color=axcolor, hovercolor='0.975')
    ax2 = plt.axes([0.1, 0.325, 0.5, 0.05])

    ax2 = plt.axes([0, 0, 0.5, 0.05])
    ax2.patch.set_alpha(0.5) 
    Slider(ax2, 'Reset2', facecolor='red', alpha=0.5, valmin=0, valmax=10)
    ax2.patch.alpha = 0.5 

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

class Annotate(object):
    def __init__(self):
        self.ax = plt.gca()
        self.rect = Rectangle((0,0), 1, 1)
        self.x0 = None
        self.y0 = None
        self.x1 = None
        self.y1 = None
        self.ax.add_patch(self.rect)
        self.ax.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.ax.figure.canvas.mpl_connect('button_release_event', self.on_release)

    def on_press(self, event):
        print ('press')
        self.x0 = event.xdata
        self.y0 = event.ydata

    def on_release(self, event):
        print ('release')
        self.x1 = event.xdata
        self.y1 = event.ydata
        self.rect.set_width(self.x1 - self.x0)
        self.rect.set_height(self.y1 - self.y0)
        self.rect.set_xy((self.x0, self.y0))
        self.ax.figure.canvas.draw()


class DraggableRectangle(object):
    def __init__(self, rect):
        self.rect = rect
        self.press = None

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return

        contains, attrd = self.rect.contains(event)
        if not contains: return
        log.info('event contains', self.rect.xy)
        x0, y0 = self.rect.xy
        self.press = x0, y0, event.xdata, event.ydata

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.press is None: return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        #log.info('x0=%f, xpress=%f, event.xdata=%f, dx=%f, x0+dx=%f' %
        #      (x0, xpress, event.xdata, dx, x0+dx))
        self.rect.set_x(x0+dx)
        self.rect.set_y(y0+dy)

        self.rect.figure.canvas.draw()


    def on_release(self, event):
        'on release we reset the press data'
        self.press = None
        self.rect.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)

        
    #a = Annotate()
    #plt.show()    

    #fig = plt.figure()
    #ax = fig.add_subplot(111)
    #rects = ax.bar(range(10), 20*np.random.rand(10))
    #drs = []
    #for rect in rects:
    #    dr = DraggableRectangle(rect)
    #    dr.connect()
    #    drs.append(dr)
    #
    #plt.show()


class DraggableRectangleFast(object):
    lock = None  # only one can be animated at a time
    def __init__(self, rect):
        self.rect = rect
        self.press = None
        self.background = None

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return
        if DraggableRectangle.lock is not None: return
        contains, attrd = self.rect.contains(event)
        if not contains: return
        log.info('event contains', self.rect.xy)
        x0, y0 = self.rect.xy
        self.press = x0, y0, event.xdata, event.ydata
        DraggableRectangle.lock = self

        # draw everything but the selected rectangle and store the pixel buffer
        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        self.rect.set_animated(True)
        canvas.draw()
        self.background = canvas.copy_from_bbox(self.rect.axes.bbox)

        # now redraw just the rectangle
        axes.draw_artist(self.rect)

        # and blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if DraggableRectangle.lock is not self:
            return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        self.rect.set_x(x0+dx)
        self.rect.set_y(y0+dy)

        canvas = self.rect.figure.canvas
        axes = self.rect.axes
        # restore the background region
        canvas.restore_region(self.background)

        # redraw just the current rectangle
        axes.draw_artist(self.rect)

        # blit just the redrawn area
        canvas.blit(axes.bbox)

    def on_release(self, event):
        'on release we reset the press data'
        if DraggableRectangle.lock is not self:
            return

        self.press = None
        DraggableRectangle.lock = None

        # turn off the rect animation property and reset the background
        self.rect.set_animated(False)
        self.background = None

        # redraw the full figure
        self.rect.figure.canvas.draw()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)    



import os
from vipy.video import VideoCategory
from vipy.util import remkdir, filetail, isvideo, isinstalled
import vipy.downloader

URL = 'http://serre-lab.clps.brown.edu/wp-content/uploads/2013/10/hmdb51_org.rar'
SHA1 = None


class HMDB(object):
    def __init__(self, datadir):
        """Human motion dataset, provide a datadir='/path/to/store/hmdb' """
        self.datadir = remkdir(datadir)

        if not os.path.exists(os.path.join(datadir, filetail(URL))):
            self.download()
        if not len(vipy.util.dirlist(datadir)) > 1:
            self._unpack(os.path.join(self.datadir, filetail(URL)), self.datadir)
            
    def __repr__(self):
        return str('<vipy.data.hmdb: "%s">' % self.datadir)

    def download(self):
        vipy.downloader.download(URL, os.path.join(self.datadir, filetail(URL)))
        self._unpack(os.path.join(self.datadir, filetail(URL)), self.datadir)

    def dataset(self):
        """Return a list of VideoCategory objects"""
        vidlist = []
        for (idx_category, category) in enumerate(os.listdir(self.datadir)):
            if os.path.isdir(os.path.join(self.datadir, category)):
                for (idx_video, filename) in enumerate(os.listdir(os.path.join(self.datadir, category))):
                    if isvideo(filename):
                        vidlist.append(VideoCategory(filename=os.path.join(self.datadir, category, filename), category=category))
        return vidlist

    def _unpack(self, rarfile, outdir):
        """Require unrar on command line"""
        if not isinstalled('unrar'):
            raise ValueError('Unpacking requires the unrar utility on the command line.  On Ubuntu: "sudo apt install unrar"')
        os.system('unrar e %s %s' % (rarfile, outdir))
        for (idx_category, rarfile) in enumerate(os.listdir(outdir)):
            (category, ext) = os.path.splitext(rarfile)
            if not os.path.isdir(os.path.join(outdir,category)):
                os.mkdir(os.path.join(outdir, category))
                os.mkdir(os.path.join(outdir, category, 'export'))
                cmd = 'unrar e %s %s' % (os.path.join(outdir,rarfile), os.path.join(outdir,category))
                os.system(cmd)



import vipy
import numpy as np
import copy
import math

vipy.util.try_import('torch')
import torch



class GaussianPyramid():
    """vipy.pyramid.GaussianPyramid() class"""
    def __init__(self, im=None, tensor=None):
        assert im is not None or tensor is not None
        assert im is None or isinstance(im, vipy.image.Image)
        assert tensor is None or (torch.is_tensor(tensor) and tensor.ndim == 4)
        
        g = ((1.0/np.sqrt(2*np.pi))*np.exp(-0.5*(np.array([-2,-1,0,1,2])**2))).astype(np.float32)
        G = torch.from_numpy(np.outer(g,g)).repeat( (3,1,1,1) )
        
        self._G = G
        self._band = []
        self._pad = torch.nn.ReflectionPad2d(2)
        
        x = im.float().torch(order='NCHW') if im is not None else tensor
        for k in range(int(np.log2(min(x.shape[2], x.shape[3]))-2)):
            y = torch.nn.functional.conv2d(self._pad(x), self._G, groups=3)  # anti-aliasing
            self._band.append(y)  # lowpass
            x = y[:,:,::2,::2]  # downsample
        self._band.append(x)  # lowpass

    def __len__(self):
        return len(self._band)

    def __iter__(self):
        for k in range(len(self)):
            yield self[k]
    
    def __getitem__(self, k):
        assert k>=0 and k<len(self)
        return vipy.image.Image.fromtorch(self._band[k])

    def band(self, k):
        return self[k]

    def show(self, mindim=256):
        return vipy.visualize.montage([im.maxsquare() for im in self], mindim, mindim).show()

    
class LaplacianPyramid():
    """vipy.pyramid.LaplacianPyramid() class"""    
    def __init__(self, im, pad='zero'):
        
        g = (1.0/np.sqrt(2*np.pi))*np.exp(-0.5*(np.array([-2,-1,0,1,2])**2))
        G = torch.from_numpy(np.outer(g,g).astype(np.float32))

        self._G = G.repeat( (3,1,1,1) )
        self._Ce = torch.sum(G[::2,::2])  # filter coefficient sum, even elements only
        self._Co = torch.sum(G[1::2,1::2]) # filter coefficient sum, odd elements only
        self._band = []
        self._im = im
        
        if pad == 'zero':
            self._pad = torch.nn.ZeroPad2d(2)  # introduces lowpass corner boundary artifact on reconstruction
            self._gain = 1.6  # rescale to approximately correct boundary artifact
        elif pad == 'reflect':
            self._pad = torch.nn.ReflectionPad2d(2)  # introduces lowpass gain boundary artifact on reconstruction
            self._gain = 1.4  # rescale to approximately correct boundary artifact
        else:
            raise ValueError('unknown padding "%s" - must be ["zero", "reflect"]' % pad)

        if isinstance(im, vipy.image.Image):
            x = im.float().torch(order='NCHW')
            for k in range(int(np.log2(im.mindim())-1)):
                y = torch.nn.functional.conv2d(self._pad(x), self._G, groups=3)  # anti-aliasing
                self._band.append(x-y)  # bandpass
                x = y[:,:,::2,::2]  # even downsample
            self._band.append(x)  # lowpass
        elif isinstance(im, list):
            assert all([torch.is_tensor(b) for b in im])
            self._band = im
            self._im = self[0].clone().zeros()  # no source image available
        else:
            raise ValueError('invalid input')

    def __repr__(self):
        return '<vipy.pyramid.LaplacianPyramid: scales=%d, channels=%d, height=%d, width=%d>' % (self.scales(), self.channels(), self.height(), self.width())
    
    def __len__(self):
        return len(self._band)

    def __iter__(self):
        for k in range(len(self)):
            yield self[k]
            
    def __getitem__(self, k):
        assert k>=0 and k<len(self)
        return vipy.image.Image.fromtorch(self._band[k])

    def band(self, k):
        return self[k]

    def reconstruct(self):
        x = self._band[-1]
        for b in reversed(self._band[0:-1]):
            xz = torch.zeros(1, b.shape[1], b.shape[2], b.shape[3])
            xz[:,:,::2,::2] = x  # odd zero interpolate (even signal)
            xu = torch.nn.functional.conv2d(self._pad(xz), (1.0/self._Ce)*self._G, groups=3)  # upsample, rescale using sum of non-zeroed kernel elements for even locations
            xu[:,:,1::2,1::2] *= (self._Ce/self._Co)  # upsample, rescale using sum of non-zeroed kernel elements for odd locations
            x = xu + b  # reconstruction
        im = vipy.image.Image.fromtorch((x * self._gain).clamp(0, 255))  # final rescale due to padding boundary artifact (can also use im.mat2gray)
        return im.array(im.numpy().astype(np.uint8)).colorspace('rgb')

    def show(self, mindim=256):
        return vipy.visualize.montage([im.maxsquare().resize(mindim, mindim, interp='nearest') for im in self], mindim, mindim).show()

    def height(self):
        return self._im.height()
    def width(self):
        return self._im.width()
    def channels(self):
        return self._im.channels()
    def scales(self):
        return len(self)
    
    def tensor(self, interp='nearest'):
        return torch.stack([im.resize(self.height(), self.width(), interp=interp).torch(order='CHW') for im in self])  # scales() x channels() x height() x width()

    @staticmethod
    def fromtensor(bands):
        """Convert a S*CxHxW torch tensor back to LaplacianPyramid"""
        assert torch.is_tensor(bands) and bands.ndim == 4
        (S,C,H,W) = bands.shape
        return LaplacianPyramid([vipy.image.Image.fromtorch(b).resize(H//(2**i), W//(2**i), interp='nearest').torch(order='NCHW') for (i,b) in enumerate(bands)])

    
class Foveation(LaplacianPyramid):
    def __init__(self, im, mode='log-circle', s=None):        
        super().__init__(im)

        (H,W) = (im.height(), im.width())
        allowable_modes = ['gaussian', 'linear-circle', 'linear-square', 'log-circle']
        if mode == 'gaussian':
            s = s*(W/2) if s is not None else W/2            
            G = vipy.math.gaussian2d([W,H], [s,s], 2*H, 2*W)
            M = np.repeat(G[:,:,np.newaxis], 3, axis=2)
            thresholds = np.arange(0, float(np.max(G)), np.max(G)/len(self))
            masks = [vipy.image.Image(array=255*np.array(M>=t).astype(np.uint8)).blur(16).mat2gray(min=0) for t in thresholds]
        elif mode == 'linear-circle':
            s = s*2.0 if s is not None else 2.0
            masks = [vipy.calibration.imcircle(W, H, s*(d/2), 2*W, 2*H, 3).rgb().blur(16).mat2gray(min=0) for d in np.arange(max(H,W), 0, -max(H,W)/len(self))]
        elif mode == 'log-circle':
            s = s*0.125 if s is not None else 0.125                        
            masks = [vipy.calibration.imcircle(W, H, (s*(d/2))**2, 2*W, 2*H, 3).rgb().blur(16).mat2gray(min=0) for d in np.arange(max(H,W), 0, -max(H,W)/len(self))]
        elif mode == 'linear-square':
            s = s*2.0 if s is not None else 2.0                        
            masks = [vipy.image.Image(array=vipy.calibration.square(W,H,s*(d/2),2*W,2*H,3).astype(np.float32), colorspace='float') for d in np.arange(max(H,W), 0, -max(H,W)/len(self))]
        else:
            raise ValueError('invalid mode "%s" - must be in %s' % (mode, str(allowable_modes)))
        self._immasks = masks
        self._masks = [m.torch(order='NCHW') for m in masks]

    def __call__(self, tx=0, ty=0, sx=1.0, sy=1.0):
        (C, H,W) = (self.channels(), self.height(), self.width())        
        theta = torch.FloatTensor([[sx,0,tx],[0,sy,ty]]).repeat( (1, 1, 1) )
        G = torch.nn.functional.affine_grid(theta, (1, C, H, W), align_corners=False)
        blend = [GaussianPyramid(tensor=torch.nn.functional.grid_sample(m, G, align_corners=False)) for m in self._masks]        

        pyr = copy.deepcopy(self)
        pyr._band[:-1] = [torch.mul(w[k].torch(order='NCHW'), b) for (k, (w,b)) in enumerate(zip(reversed(blend), pyr._band[:-1]))]
        return pyr.reconstruct()
        
    def foveate(self, tx=0, ty=0):
        """Foveate the input image at location (tx, ty) in scaled image coordinates where (0,0) is the center and (1,1) is the upper left"""        
        return self.__call__(tx=tx, ty=ty, sx=1.0, sy=1.0)

    def visualize(self):
        """Show the fovea density"""
        imgmask = self._immasks[0].clone().mat2gray().numpy()
        for (im,c) in zip(self._immasks, range(0, 255, len(self))):
            img = im.clone().mat2gray(min=0).numpy()
            imgmask += c*img
        return vipy.image.Image(array=imgmask)

    
class SteerablePyramid():
    def __init__(self, im):
        vipy.util.try_import('pyrtools')
        import pyrtools
        
        assert isinstance(im, vipy.image.Image)
        assert im.mindim() >= 32        
        self._channels = [pyrtools.pyramids.SteerablePyramidFreq(imc.load().array().astype(float), height='auto', order=3) for imc in im.channel()]                
        
    @property
    def num_scales(self):
        return self._channels[0].num_scales

    @property
    def num_orientations(self):
        return self._channels[0].num_orientations

    @property
    def num_channels(self):
        return len(self._channels)           
        
    def bandpass(self, channel):
        return [self._channels[channel].pyr_coeffs[(s,o)] for s in range(self.num_scales) for o in range(self.num_orientations)]
    
    def lowpass(self, channel):
        return self._channels[channel].pyr_coeffs['residual_lowpass']

    def highpass(self, channel):
        return self._channels[channel].pyr_coeffs['residual_highpass']        
    
    def synthesis(self):        
        return vipy.image.Image(array=np.stack([pyr.recon_pyr() for pyr in self._channels], axis=2).astype(np.float32)).mat2gray()

    def multichannel(self):
        """a multichannel image is an image of the same shape as the input, but with channels from pyramid decomposition.  Coefficients are resized using bilinear interpolation."""
        (H,W) = self.highpass(0).shape
        resizer = lambda x: np.array(torch.nn.functional.interpolate(torch.tensor(x).view(1,1,x.shape[0],x.shape[1]), size=(H,W), mode='bilinear')).squeeze()            
        return vipy.image.Image(array=np.stack([resizer(x) for c in range(self.num_channels) for x in self.bandpass(c)+[self.lowpass(c),self.highpass(c)]], axis=2).astype(np.float32))

    def montage(self):
        """scales by row, orientations by col, channels merged back into color image, last image is lowpass"""
        imlist = [vipy.image.Image(array=np.stack([self._channels[c].pyr_coeffs[(s,o)] for c in range(self.num_channels)], axis=2).astype(np.float32)) for s in range(self.num_scales) for o in range(self.num_orientations)]
        imlist += [vipy.image.Image(array=np.stack([self.lowpass(c) for c in range(self.num_channels)], axis=2).astype(np.float32))]
        imlist += [vipy.image.Image(array=np.stack([self.highpass(c) for c in range(self.num_channels)], axis=2).astype(np.float32))]        
        return vipy.visualize.montage([im.mat2gray().rgb() for im in imlist], gridrows=self.num_scales+1, gridcols=self.num_orientations)

    

class BatchSteerablePyramid():
    def __init__(self, height, width, channels, device='cpu', orientations=4):
        vipy.util.try_import('plenoptic')
        import plenoptic as po

        assert height >= 32 and width >= 32        
        self._pyr = po.simulate.SteerablePyramidFreq(height='auto', image_shape=[height, width], twidth=1, order=orientations-1, downsample=False, is_complex=True).to(device)
        self._pyr.eval()
        self._imheight = height
        self._imwidth = width
        self._imchannels = channels
        self._device = device
        self._pyr_info = None
        self._tensor = None
        
    def device(self, d=None):
        if d is not None:
            self._device = d  # check device in self._pyr._buffers['lo0mask'], or just store here
            self._pyr = self._pyr.cpu() if d == 'cpu' else self._pyr.to(d)
            return self
        return self._device
                
    @property
    def num_scales(self):
        return self._pyr.num_scales

    @property
    def num_orientations(self):
        return self._pyr.num_orientations

    @property
    def num_channels(self):
        return (self._pyr.num_orientations*self._pyr.num_scales + 2)*self._imchannels

    def to(self, dev):
        return self.device(dev)    

    def forward(self, x):
        assert (torch.is_tensor(x) and x.ndim == 4) or isinstance(x, vipy.image.Image) or (isinstance(x, (list, tuple)) and all([isinstance(im, vipy.image.Image) for im in x]))

        x = torch.stack([im.torch('CHW') for im in vipy.util.tolist(x)]) if not torch.is_tensor(x) else x
        assert x.shape[2] == self._imheight and x.shape[3] == self._imwidth, "wrong input shape"
        assert x.shape[1] == self._imchannels, "wrong input channels"

        with torch.no_grad():        
            (self._tensor, self._pyr_info) =  self._pyr.convert_pyr_to_tensor(self._pyr.forward(x.to(self.device())))
        return self

    def forwarded(self):
        return self._tensor is not None
    
    def tensor(self):
        """a pyramid tensor is an NxCxHxW tensor (same shape as the input), but with channels from complex (even, odd) pyramid coefficients
        The first channel will be the residual highpass and the last will be the residual lowpass. Each band is then a separate channel in (scale, orientation) order

        input is NxCxHxW tensor
        """
        assert self.forwarded() 
        return self._tensor

    def band(self, scale, orientation):
        assert scale <= self.num_scales and orientation <= self.num_orientations
        return self.tensor()[:, 1+scale*self.num_orientations + orientation,:,:]

    def magnitude(self):
        """return magnitude component of complex steerable pyramid"""
        return self.tensor().abs()

    def phase(self):
        """return phase component of complex steerable pyramid"""
        return self.tensor().angle()        
    
    def montage(self, x):
        """return a montage visualization of the pyramid, scales by row, orientations by col, channels merged back into color image, last row is highpass and lowpass"""
        assert self.forwarded()
        assert torch.is_tensor(x) and x.ndim == 4 and x.shape == self._tensor.shape
        
        (C,N) = (x.shape[1], x.shape[1]//self._imchannels)
        return vipy.visualize.montage([vipy.image.Image.fromtorch(x[0,j::N,:,:]).mat2gray().rgb() for j in list(range(1,N-1))+[0,N-1]],
                                       gridrows=self.num_scales+1, gridcols=self.num_orientations)                

        
    def synthesis(self, x):
        """Generate a synthesis of the multichannel tensor x"""
        assert self.forwarded()
        assert torch.is_tensor(x) and x.ndim == 4 and x.shape == self._tensor.shape

        img = self._pyr.recon_pyr(self._pyr.convert_tensor_to_pyr(x, pyr_keys=self._pyr_info[2], num_channels=self._pyr_info[0], split_complex=self._pyr_info[1]))
        return [vipy.image.Image.fromtorch(t) for t in img]  # one image per batch element

    
    def phase_congruency(self, eps=1E-6):
        (M,P) = (self.magnitude()[:,1:-1,:,:], self.phase()[:,1:-1,:,:])
        return torch.sum( M * torch.cos(P - ((M*P).sum(dim=1, keepdims=True) / (eps + M.sum(dim=1, keepdims=True)))), dim=1, keepdims=True) / (eps + torch.sum(M, dim=1, keepdims=True))

    def zero_crossing(self, eps=1E-6):
        (M,P) = (self.magnitude()[:,1:-1,:,:], self.phase()[:,1:-1,:,:])
        return torch.sum( M*torch.nn.functional.relu(torch.cos(P)), dim=1, keepdims=True)
    



import urllib
import os
import re
import random
import math
from vipy.util import try_import, isurl, remkdir, temphtml, tempjson, filetail, readjson, isjsonfile, toextension
import tempfile
import numpy as np
from vipy.downloader import common_user_agents
from vipy.globals import log
from datetime import datetime
import time
import webbrowser
import html
            

def googlesearch(tag):
    """Return a list of image URLs from google image search associated with the provided tag"""
    raise ValueError('no longer unsupported')
    
    try_import('bs4', 'BeautifulSoup4');  
    from bs4 import BeautifulSoup
    
    url = 'https://www.google.com/search?tbm=isch&q=%s' % tag.replace(' ','+')
    user_agent = random.choice(common_user_agents)
    headers = {'User-Agent':user_agent}
    search_request = urllib.request.Request(url,None,headers)
    search_results = urllib.request.urlopen(search_request)
    search_data = str(search_results.read())

    soup = BeautifulSoup(search_data, features="html.parser")
    links = soup.find_all('a')    
    urls = [tag.get('href',None) for tag in links if tag.get('href',None) is not None]
    urls = [u.replace('/url?q=http','http') if u.startswith('/url?q=http') else u for u in urls]
    urls = [u for u in urls if u.startswith('http') and not u.startswith('https://www.google.com')]
    return urls


def basic_level_categories():
    """Return a list of nouns from wordnet that can be used as an initial list of basic level object categories"""
    try_import('nltk'); import nltk
    nltkdir = remkdir(os.path.join(os.environ['VIPY_CACHE'], 'nltk')) if 'VIPY_CACHE' in os.environ else tempfile.gettempdir()
    os.environ['NLTK_DATA'] = nltkdir
    log.info('[vipy.annotation.basic_level_categories]: Downloading wordnet to "%s"' % nltkdir)
    nltk.download('wordnet', nltkdir)
    nltk.data.path.append(nltkdir)
    
    from nltk.corpus import wordnet
    nouns = []
    allowed_lexnames = ['noun.animal', 'noun.artifact', 'noun.body', 'noun.food', 'noun.object', 'noun.plant']
    for synset in list(wordnet.all_synsets('n')):
        if synset.lexname() in allowed_lexnames:
            nouns.append(str(synset.lemmas()[0].name()).lower())
    nouns.sort()
    return nouns

def nouns():
    """Return nouns from wordnet as a dictionary of imagenet 'wordnet id' (e.g. n + offset) keys to a list of synset names""" 
    try_import('nltk'); import nltk
    nltkdir = remkdir(os.path.join(os.environ['VIPY_CACHE'], 'nltk')) if 'VIPY_CACHE' in os.environ else tempfile.gettempdir()
    os.environ['NLTK_DATA'] = nltkdir
    log.info('[vipy.annotation.nouns]: Downloading wordnet to "%s"' % nltkdir)
    nltk.download('wordnet', nltkdir)
    nltk.data.path.append(nltkdir)

    from nltk.corpus import wordnet
    nouns = {'n%d' % s.offset():s.lemma_names() for s in wordnet.all_synsets('n')}
    return dict(sorted(nouns.items(), key=lambda x: x[0]))
    

def verbs():
    """Return a list of verbs from verbnet that can be used to define a set of activities"""
    try_import('nltk'); import nltk
    nltkdir = remkdir(os.path.join(os.environ['VIPY_CACHE'], 'nltk')) if 'VIPY_CACHE' in os.environ else tempfile.gettempdir()
    os.environ['NLTK_DATA'] = nltkdir
    log.info('[vipy.annotation.verbs]: Downloading verbnet to "%s"' % nltkdir)    
    nltk.download('verbnet', nltkdir)
    nltk.data.path.append(nltkdir)
    from nltk.corpus import verbnet
    return verbnet.lemmas()


class Selector():
    """HTML+javascript based UI for selecting images from a group.  

    * Given a list of lists of `vipy.image.Image` objects, display each image in a standalone HTML file.  
    * Allow the user to click on desirable images to select them.  
    * When done, the user selects the Download JSON button to export the selected elements.
    * The selected elements are stored in a JSON file which is exported back into this object to return only the selected images

    ```python
    imlist = [[vipy.image.RandomScene(128,128) for k in range(10)] for j in range(5)]
    s = vipy.annotation.Selector(imlist)
    s.html(show=True)  # annotate locally in browser by selecting images to keep, then download the resulting jsonfile
    selected_imlist = s.fromjson('/path/to/downloaded.json')  # returns only those images selected in the UI
    ```
    """
       
    def __init__(self, imlist):
        self._imlist = imlist

    def __iter__(self):
        for im in self._imlist:
            yield im
            
    def html(self, jsonfile=None, htmlfile=None, imdict=None, title='Image Selector', maxwidth=4096, default_opacity=0.5, selected_opacity=1.0, show=False, description="Select the images to keep from each group, download the JSON, then call vipy.annotation.Select.fromjson()", mindim=None):
        """Given a list of tuples of `vipy.image.Image` objects, create a standalone HTML file that will allow the user to select individual images in the group.  Selected images are output to jsonfile"""
        htmlfile = temphtml() if htmlfile is None else htmlfile
        jsonfile = filetail(toextension(htmlfile, '.json')) if jsonfile is None else jsonfile
        imlist = self._imlist
        
        f = open(htmlfile,'w')
        f.write('<!DOCTYPE html>\n')
        f.write('<!--\n    Visym Labs\n    vipy.annotation.Selector (https://visym.github.io/vipy)\n    Generated: %s\n-->\n' % str(datetime.now()))            
        f.write("<head>\n");
        f.write('  <meta charset="UTF-8">\n')
        f.write("  <title>%s</title>\n" % title)
        f.write('  <link href="https://use.fontawesome.com/releases/v5.13.0/css/all.css" rel="stylesheet">\n')
        f.write("</head>\n")        
        f.write('<html>\n')
        f.write('<body>\n')
        f.write('  <script type="text/javascript">\n')
        f.write('    var selected = {};\n');
        f.write('    function set_unselected(x) { x.style.opacity=%s; x.setAttribute("selected","off"); selected[x.id]=false;};\n' % (default_opacity))
        f.write('    function set_selected(x) { x.style.opacity=%s; x.setAttribute("selected","on"); selected[x.id]=true;};\n' % (selected_opacity))
        f.write('    function toggle(x) { if (x.getAttribute("selected") == "off") { set_selected(x); } else { set_unselected(x); }; }\n')
        f.write('    function row_selected(k, n) { for (let i = 0; i < n; i++) { set_selected( document.getElementById(`(${k},${i})`));};}\n')
        f.write('    function row_unselected(k, n) { for (let i = 0; i < n; i++) { set_unselected( document.getElementById(`(${k},${i})`));};}\n')
        f.write('    function toggle_row(x,k,n) { if (x.getAttribute("selected") == "off") { row_selected(k,n); x.setAttribute("selected","on"); } else { row_unselected(k,n); x.setAttribute("selected","off");}; }\n')        
        f.write('    function exportJSON(x) { var data = "text/json;charset=utf-8," + encodeURIComponent(JSON.stringify(selected)); x.setAttribute("href", "data:"+data); x.setAttribute("download", "%s"); }\n' % jsonfile)
        f.write('    function set_transparent(x) { x.style.opacity=0.5; };\n')
        f.write('    function set_opaque(x) { x.style.opacity=1.0; };\n')        
        f.write('  </script>\n')    
        f.write('<div id="container" style="width:%dpx">\n' % maxwidth)
        f.write('<div id="header">\n')
        f.write('<h1 style="margin-bottom:0;">%s</h1><br>\n' % title)
        localtime = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(time.time()))
        f.write('HTML generated on %s<br>\n\n' % localtime)
        f.write('Number of image groups: %d<br>\n' % len(imlist))
        f.write('Number of images: %d<br><br>\n' % len([im for iml in imlist for im in iml]))
        f.write('%s<br><br>\n' % description)
        f.write('<a onclick="exportJSON(this);" class="btn"><button>Download JSON</button></a>\n') 
        f.write('</div>\n')
        f.write('<br>\n')
        f.write('<hr>\n')
        f.write('<div style="float:left;">\n')
    
        # Generate embedded images and html
        for (i, imsrclist) in enumerate(imlist):
            f.write('<p>\n</p>\n')
            if imdict is not None:
                for (k,v) in imdict[i].items():
                    f.write('<b>%s</b>: %s<br>\n' % (html.escape(str(k)), html.escape(str(v))))
                f.write('<p>\n</p>\n')
                
            for (j, im) in enumerate(imsrclist):                    
                im = im if mindim is None else im.clone().mindim(mindim)
                f.write(im.html(id='(%d,%d)' % (i,j), attributes={'loading':'lazy', 'onclick':"toggle(this)", 'selected':'off', 'style':"opacity:%f" % default_opacity}))   # base-64 encoded image with img tag
            f.write('<br><i title="Toggle Select All" onclick="toggle_row(this,%d,%d);" selected="on" onmouseover="set_opaque(this)" onmouseout="set_transparent(this)" class="fas fa-recycle fa-3x" id="row-%d" style="opacity:0.5;margin-right: 10px;"></i>' % (i,len(imsrclist),i))
            
            f.write('<p>\n</p>\n')
            f.write('<br>\n')
            f.write('<hr>\n')
        f.write('</div>\n')
        f.write('</body>\n')
        f.write('</html>\n')
        f.close()

        if show:
            webbrowser.open('file:///%s' % htmlfile)  # chrome only, will not work in safari
        return htmlfile
    
    def fromjson(self, jsonfile):
        """Given the JSON file downloaded from the HTML selector, return only those that were selected"""
        assert isjsonfile(jsonfile)
        d = readjson(jsonfile)
        return [[im for (j,im) in enumerate(iml) if '(%d,%d)'%(i,j) in d and d['(%d,%d)'%(i,j)] == True] for (i,iml) in enumerate(self._imlist)]

    
class Filter(Selector):
    """HTML+javascript based UI for filtering images from a group.  

    * Given a list of lists of `vipy.image.Image` objects, display each image in a standalone HTML file.  
    * Allow the user to click on non-desirable images to remove them.  
    * When done, the user selects the Download JSON button to export the filtered elements.
    * The filtered elements are stored in a JSON file which is exported back into this object to return all but the filtered images

    ```python
    imlist = [[vipy.image.RandomScene(128,128) for k in range(10)] for j in range(5)]
    s = vipy.annotation.Filter(imlist)
    s.html(show=True)  # annotate locally in browser by selecting images to remove and download the resulting jsonfile
    filtered_imlist = s.fromjson('/path/to/downloaded.json')   # returns images with those selected in the UI removed
    ```
    """
    
    def html(self, jsonfile=None, htmlfile=None, imdict=None, title='Image Filter', maxwidth=4096, default_opacity=1.0, selected_opacity=0.5, show=False, description="Select the images to remove from each group, download the JSON, then call vipy.annotation.Filter.fromjson()", mindim=None):
        return super().html(jsonfile=jsonfile, htmlfile=htmlfile, title=title, default_opacity=default_opacity, selected_opacity=selected_opacity, show=show, description=description, mindim=mindim, imdict=imdict, maxwidth=maxwidth)

    def fromjson(self, jsonfile):    
        """Given the JSON file downloader from the HTML selector, return a filter of only those that were *not* selected"""
        assert isjsonfile(jsonfile)
        d = readjson(jsonfile)
        return [[im for (j,im) in enumerate(iml) if '(%d,%d)'%(i,j) not in d or d['(%d,%d)'%(i,j)] == False] for (i,iml) in enumerate(self._imlist)]
            
    



import os
import vipy


URL = 'https://objectnet.dev/downloads/objectnet-1.0.zip'


class Objectnet(vipy.dataset.Dataset):
    """Project: https://objectnet.dev, password set on website, must be bytes encoded (e.g. passwd=b'thepassword')"""
    def __init__(self, datadir=None, passwd='objectnetisatestset', url=URL, name='objectnet', redownload=False):

        datadir = tocache('objectnet') if datadir is None else datadir
        
        self._datadir = vipy.util.remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(url, self._datadir, md5=None, passwd=passwd)

        imlist = vipy.util.findimages(os.path.join(datadir, 'objectnet-1.0'))
        loader = lambda f: vipy.image.ImageCategory(filename=f, category=vipy.util.filebase(vipy.util.filepath(f)))
        super().__init__(imlist, id=name, loader=loader)

        open(os.path.join(self._datadir, '.complete'), 'a').close()



import os
import vipy

    
class TinyVIRAT():
    URL = 'https://www.crcv.ucf.edu/tiny-actions-challenge-cvpr2021/data/TinyVIRAT-v2.zip'
    def __init__(self, datadir):
        jsonfile = os.path.join(datadir, 'TinyVIRAT_V2','tiny_train_v2.json')
        if not os.path.exists(jsonfile):
            vipy.downloader.download_and_unpack(TinyVIRAT.URL, datadir)

        self._trainset = vipy.util.readjson(os.path.join(datadir, 'TinyVIRAT_V2','tiny_train_v2.json'))
        self._valset = vipy.util.readjson(os.path.join(datadir, 'TinyVIRAT_V2','tiny_val_v2.json'))
        self._testset = vipy.util.readjson(os.path.join(datadir, 'TinyVIRAT_V2','tiny_test_v2_public.json'))                
        self._datadir = datadir
        
    def trainset(self):
        return vipy.dataset.Dataset([vipy.video.Scene(tags=t['label'], filename=os.path.join(self._datadir, 'TinyVIRAT_V2', 'videos', 'train', t['path'])) for t in self._trainset], id='TinyVIRAT:train')

    def valset(self):
        return vipy.dataset.Dataset([vipy.video.Scene(tags=t['label'], filename=os.path.join(self._datadir, 'TinyVIRAT_V2', 'videos', 'val', t['path'])) for t in self._valset], id='TinyVIRAT:val')

    def testset(self):
        return vipy.dataset.Dataset([vipy.video.Video(filename=os.path.join(self._datadir, 'TinyVIRAT_V2', 'videos', 'test', t['path'])) for t in self._testset], id='TinyVIRAT:test')
    



import os
import vipy


URL = 'http://data.vision.ee.ethz.ch/cvl/food-101.tar.gz'
SHA1 = 'ed21dfefc61fbe39294b9441739f7ac91b343882'


class Food101(vipy.dataset.Dataset):
    """Project: https://data.vision.ee.ethz.ch/cvl/datasets_extra/food-101/"""
    def __init__(self, datadir=None, redownload=False):
        datadir = tocache('food101') if datadir is None else datadir        

        # Download
        self._datadir = vipy.util.remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(URL, self._datadir, sha1=SHA1)            

        loader = lambda f: vipy.image.ImageCategory(filename=f, category=vipy.util.filebase(vipy.util.filepath(f)))
        imlist = vipy.util.findimages(os.path.join(datadir, 'food-101'))
        super().__init__(imlist, id='food101', loader=loader)

        open(os.path.join(self._datadir, '.complete'), 'a').close()








import os
import numpy as np
from vipy.util import findpkl, toextension, filepath, filebase, jsonlist, ishtml, ispkl, filetail, temphtml, env, cache
from vipy.util import listpkl, listext, templike, tempdir, remkdir, tolist, fileext, writelist, tempcsv, to_iterable
from vipy.util import newpathroot, listjson, extlist, filefull, tempdir, groupbyasdict, try_import, shufflelist, catcher
from vipy.util import is_email_address, findjson, findimages, isimageurl, countby, chunkgen, chunkgenbysize, dividelist, chunklist
from vipy.globals import log
import random
import shutil
import copy 
import gc 
import itertools
from pathlib import Path
import functools
import concurrent.futures as cf
import vipy.parallel


class Dataset():
    """vipy.dataset.Dataset() class
    
    Common class to manipulate large sets of objects in parallel

    Args:
        - dataset [list, tuple, set, obj]: a python built-in type that supports indexing or a generic object that supports indexing and has a length
        - id [str]: an optional id of this dataset, which provides a descriptive name of the dataset
        - loader [callable]: a callable loader that will construct the object from a raw data element in dataset.  This is useful for custom deerialization or on demand transformations

    Datasets can be indexed, shuffled, iterated, minibatched, sorted, sampled, partitioned.
    Datasets constructed of vipy objects are lazy loaded, delaying loading pixels until they are needed

    ```python
    (trainset, valset, testset) = vipy.dataset.registry('mnist')

    (trainset, valset) = trainset.partition(0.9, 0.1)
    categories = trainset.set(lambda im: im.category())
    smaller = testset.take(1024)
    preprocessed = smaller.map(lambda im: im.resize(32, 32).gain(1/256))
    
    for b in preprocessed.minibatch(128):
        print(b)

    # visualize the dataset 
    (trainset, valset, testset) = vipy.dataset.registry('pascal_voc_2007')
    for im in trainset:
        im.mindim(1024).show().print(sleep=1).close()
    
    ```

    Datasets can be constructed from directories of json files or image files (`vipy.dataset.Dataset.from_directory`)
    Datasets can be constructed from a single json file containing a list of objects (`vipy.dataset.Dataset.from_json`)
    
    ..note:: that if a lambda function is provided as loader then this dataset is not serializable.  Use self.load() then serialize
    """

    __slots__ = ('_id', '_ds', '_idx', '_loader', '_type')
    def __init__(self, dataset, id=None, loader=None):
        assert loader is None or callable(loader)
        
        self._id = id
        self._ds = dataset if not isinstance(dataset, Dataset) else dataset._ds
        self._idx = None if not isinstance(dataset, Dataset) else dataset._idx   # random access on-demand
        self._loader = loader if not isinstance(dataset, Dataset) else dataset._loader  # not serializable if lambda is provided

        try:
            self._type = str(type(self._loader(dataset[0]) if self._loader else dataset[0]))  # peek at first element, cached
        except:
            self._type = None


    @classmethod
    def from_directory(cls, indir, filetype='json', id=None):
        """Recursively search indir for filetype, construct a dataset from all discovered files of that type"""
        if filetype == 'json':
            return cls([x for f in findjson(indir) for x in to_iterable(vipy.load(f))], id=id)
        elif filetype.lower() in ['jpg','jpeg','images']:
            return cls([vipy.image.Image(filename=f) for f in findimages(indir)], id=id)            
        else:
            raise ValueError('unsupported file type "%s"' % filetype)

    @classmethod
    def from_image_urls(cls, urls, id=None):
        """Construct a dataset from a list of image URLs"""
        return cls([vipy.image.Image(url=url) for url in to_iterable(urls) if isimageurl(url)], id=id)
        
    @classmethod
    def from_json(cls, jsonfile, id=None):
        return cls([x for x in to_iterable(vipy.load(jsonfile))], id=id)

    @classmethod
    def cast(cls, obj):
        return cls(obj) if not isinstance(obj, Dataset) else obj
    
    def __repr__(self):
        fields = ['id=%s' % self.id(truncated=True, maxlen=80)] if self.id() else []
        fields += ['len=%d' % self.len()] if self.len() is not None else []
        fields += ['type=%s' % self._type] if self._type else []
        return str('<vipy.dataset.Dataset: %s>' % ', '.join(fields))

    def __parallel_iter__(self, mapper=None, accepter=None, bufsize=1024):       
        assert mapper is None or callable(mapper)
        assert accepter is None or callable(accepter)

        return (self.__iter__(mapper=mapper, accepter=accepter) if vipy.globals.cf() is None else
                vipy.parallel.iter(self, mapper=mapper, bufsize=bufsize, accepter=accepter))
                        
    def __iter__(self, mapper=None, accepter=None):            
        if self.is_streaming():
            for x in self._ds:  # iterable access (faster)
                x = self._loader(x) if self._loader is not None else x                 
                x = mapper(x) if mapper is not None else x
                if accepter is None or accepter(x):
                    yield x  
        else:
            for k in range(len(self)):
                x = self[k]   # random access (slower)                
                x = mapper(x) if mapper is not None else x
                if accepter is None or accepter(x):
                    yield x                  

    def __getitem__(self, k):
        assert self.len() is not None, "dataset does not support indexing"
        
        idx = self.index()  # convert to random access on demand
        if isinstance(k, (int, np.uint64)):
            assert abs(k) < len(idx), "invalid index"
            x = self._ds[idx[int(k)]]
            x = self._loader(x) if self._loader is not None else x
            return x
        elif isinstance(k, slice):
            X = [self._ds[k] for k in idx[k.start:k.stop:k.step]]
            X = [self._loader(x) for x in X] if self._loader is not None else X
            return X
        else:
            raise ValueError('invalid slice "%s"' % type(k))            

    def raw(self):
        """Return a view of this dataset without the loader"""
        return Dataset(self._ds, loader=None)
    
    def is_streaming(self):
        return self._idx is None

    def len(self):
        return len(self._idx) if self._idx is not None else (len(self._ds) if hasattr(self._ds, '__len__') else None)

    def __len__(self):
        len = self.len()
        if len is None:
            raise ValueError('dataset has no length')
        return len
    
    def __or__(self, other):
        assert isinstance(other, Dataset)
        return Union(self, other, id=self.id())
    
    def id(self, n=None, truncated=False, maxlen=80, suffix=None):
        """Set or return the dataset id, useful for showing the name/split of the dataset in the representation string"""
        if n is None and suffix is None:
            return (self._id[0:maxlen] + ' ... ') if truncated and self._id and len(self._id)>maxlen else self._id
        elif n is None and suffix is not None:
            self._id = self._id + suffix
        elif n is not None:
            self._id = n
        return self

    def index(self, index=None, strict=False):
        """Update the index, useful for filtering of large datasets"""
        if index is not None:
            assert not strict or index is None or (len(index)>0 and len(index)<=len(self) and max(index)<len(self) and min(index)>=0)            
            self._idx = index
            return self
        if self._idx is None:
            self._idx = list(range(len(self._ds)))  # on-demand index, only if underlying dataset has known length
        return self._idx
    
        
    def clone(self, deep=False):
        """Return a copy of the dataset object"""
        if not deep:
            return copy.copy(self) 
        else:
            return copy.deepcopy(self)
    
    def shuffle(self, shuffler=None):
        """Permute elements in this dataset uniformly at random in place using the optimal shuffling strategy for the dataset structure to maximize performance.
           This method will use either Dataset.streaming_shuffler (for iterable datasets) or Dataset.uniform_shuffler (for random access datasets)
        """
        assert shuffler is None or callable(shuffler)
        shuffler = shuffler if shuffler is not None else (Dataset.streaming_shuffler if self.is_streaming() else Dataset.uniform_shuffler)
        return shuffler(self)

    def repeat(self, n):
        """Repeat the dataset n times.  If n=0, the dataset is unchanged, if n=1 the dataset is doubled in length, etc."""
        assert n>=0
        return self.index( self.index()*(n+1) )
    
    def tuple(self, mapper=None, flatten=False, reducer=None):
        """Return the dataset as a tuple, applying the optional mapper lambda on each element, applying optional flattener on sequences returned by mapper, and applying the optional reducer lambda on the final tuple, return a generator"""
        assert mapper is None or callable(mapper)
        assert reducer is None or callable(reducer)
        mapped = self.map(mapper) if mapper else self
        flattened = (y for x in mapped for y in x) if flatten else (x for x in mapped)
        reduced = reducer(flattened) if reducer else flattened
        return reduced

    def list(self, mapper=None, flatten=False):
        """Return a tuple as a list, loading into memory"""
        return self.tuple(mapper, flatten, reducer=list)

    def set(self, mapper=None, flatten=False):
        """Return the dataset as a set.  Mapper must be a lambda function that returns a hashable type"""
        return self.tuple(mapper=mapper, reducer=set, flatten=flatten)        

    def frequency(self, f):
        """Frequency counts for which lamba returns the same value"""
        return countby(self.tuple(mapper=f))

    def count(self, f):
        """Counts for each element for which lamba returns true.  
        
        Args:
            f: [lambda] if provided, count the number of elements that return true.  

        Returns:
            A length of elements that satisfy f(v) = True [if f is not None]
        """
        return len(self.tuple(f, reducer=lambda X: [x for x in X if x is True]))

    def countby(self, f):
        return self.frequency(f)
    
    def filter(self, f):
        """In place filter with lambda function f, keeping those elements obj in-place where f(obj) evaluates true.  Callable should return bool"""
        assert callable(f)
        return self.index( [i for (b,i) in zip(self.localmap(f), self.index()) if b] )
    
    def take(self, n, inplace=False):
        """Randomly Take n elements from the dataset, and return a dataset (in-place or cloned)."""
        assert isinstance(n, int) and n>0
        D = self.clone() if not inplace else self
        return D.index( list(random.sample(D.index(), n)) ) 


    def groupby(self, f):
        """Group the dataset according to the callable f, returning dictionary of grouped datasets."""
        assert callable(f)        
        return {k:self.clone().index([x[1] for x in v]).id('%s:%s' % (self.id(),str(k))) for (k,v) in itertools.groupby(enumerate(self.sort(f).index()), lambda x: f(self[x[0]]))}

    def takeby(self, f, n):
        """Filter the dataset according to the callable f, take n from each group and return a dataset.  Callable should return bool"""
        return self.clone().filter(f).take(n)

    def takelist(self, n):
        """Take n elements and return list.  The elements are loaded and not cloned."""
        return self.take(n).list()

    def takeone(self):
        """Randomly take one element from the dataset and return a singleton"""
        return self[random.randint(0, len(self)-1)]

    def sample(self):
        """Return a single element sampled uniformly at random"""
        return self.takeone()
    
    def take_fraction(self, p, inplace=False):
        """Randomly take a percentage of the dataset, returning a clone or in-place"""
        assert p>=0 and p<=1, "invalid fraction '%s'" % p
        return self.take(n=int(len(self)*p), inplace=inplace)

    def inverse_frequency(self, f):
        """Return the inverse frequency of elements grouped by the callable f.  Returns a dictionary of the callable output to inverse frequency """
        attributes = self.set(f)
        frequency = self.frequency(f)
        return {a:(1/len(attributes))*(len(self)/frequency[a]) for a in attributes}  # (normalized) inverse frequency weight
    
    def load(self):
        """Cache the entire dataset into memory"""
        return Dataset([x for x in self], id=self.id())
    
    def chunk(self, n):
        """Yield n chunks as list.  Last chunk will be ragged."""
        for (k,c) in enumerate(chunkgen(self, n)):
            yield list(c)

    def batch(self, n):
        """Yield batches of size n as datasets.  Last batch will be ragged.  Batches are not loaded.  Batches have appended id equal to the zero-indexed batch order"""
        for (k,b) in enumerate(itertools.batched(self, n)):
            yield Dataset(b).id('%s:%d' % (self.id() if self.id() else '', k))
                                
    def minibatch(self, n, ragged=True, loader=None, bufsize=1024, accepter=None):
        """Yield preprocessed minibatches of size n of this dataset.

        To yield chunks of this dataset, suitable for minibatch training/testing

        ```python
        D = vipy.dataset.Dataset(...)
        for b in D.minibatch(n):
           print(b)
        ```
        
        To perform minibatch image downloading in parallel across four processes with the context manager:

        ```python
        D = vipy.dataset.registry('yfcc100m_url:train').take(128)
        with vipy.globals.parallel(4):
            for b in D.minibatch(16, loader=vipy.image.Transform.download, accepter=lambda im: im.is_downloaded()):
                print(b)  # complete minibatch that passed accepter
        ```

        Args:
            n [int]: The size of the minibatch
            ragged [bool]: If ragged=true, then the last chunk will be ragged with len(chunk)<n, else skipped
            bufsize [int]:  The size of the buffer used in parallel processing of elements.  Useful for parallel loading
            accepter [callable]:  A callable that returns true|false on an element, where only elements that return true are included in the minibatch.  useful for parallel loading of elements that may fail to download

        Returns:        
            Iterator over `vipy.dataset.Dataset` elements of length n.  Minibatches will be yielded loaded and preprocessed (processing done concurrently if vipy.parallel.executor() is initialized)

        ..note:: The distributed iterator appends the minibatch index to the minibatch.id().  
        ..note:: If there exists a vipy.parallel.exeuctor(), then loading and preprocessing will be performed concurrently

        """
        
        for (k,b) in enumerate(itertools.batched(self.__parallel_iter__(mapper=loader, accepter=accepter, bufsize=max(bufsize,n)), n)):                
            if ragged or len(b) == n:
                yield Dataset(b).id('%s:%d' % (self.id() if self.id() else '', k))                    
                    
                        
    def shift(self, m):
        """Circular shift the dataset m elements to the left, so that self[k+m] == self.shift(m)[k].  Circular shift for boundary handling so that self.shift(m)[-1] == self[m-1]"""
        return self.clone().index(self.index()[m:] + self.index()[0:m])

    def slice(self, start=0, stop=-1, step=1):
        """Slice the dataset to contain elements defined by slice(start, stop, step)"""
        return self.clone().index(self.index()[start:stop:step])
        
    def truncate(self, m):
        """Truncate the dataset to contain the first m elements only"""
        return self.slice(stop=m)
    
    def pipeline(self, n, m, ragged=True, prepad=True, postpad=True):
        """Yield pipelined minibatches of size n with pipeline length m.

        A pipelined minibatch is a tuple (head, tail) such that (head, tail) are minibatches at different indexes in the dataset.  
        Head corresponds to the current minibatch and tail corresponds to the minibatch left shifted by (m-1) minibatches.

        This structure is useful for yielding datasets for pipelined training where head contains the minibatch that will complete pipeline training on this iteration, and tail contains the 
        next minibatch to be inserted into the pipeline on this iteration.
        
        ```python
        D = vipy.dataset.Dataset(...)
        for (head, tail) in D.pipeline(n, m, prepad=False, postpad=False):
            assert head == D[0:m]
            assert tail == D[n*(m-1): n*(m-1)+n]

        Args:
            n [int]: The size of each minibatch
            m [int]:  The pipeline length in minibatches
            ragged [bool]: If ragged=true, then the last chunk will be ragged with len(chunk)<n, else skipped
            prepad: If true, yield (head, tail) == (None, batch) when filling the pipeline
            postpad: If true, yield (head, tail) == (batch, None) when flushing the pipeline
        
        Returns:        
            Iterator over tuples (head,tail) of `vipy.dataset.Dataset` elements of length n where tail is left shifted by n*(m-1) elements. 
        
        .. note::  The distributed iterator is not order preserving over minibatches and yields minibatches as completed, however the tuple (head, tail) is order preserving within the pipeline
        .. note:: If there exists a vipy.parallel.executor(), then loading and preprocessing will be performed concurrently
        
        """
        pipeline = [] 
        for (k,b) in enumerate(self.minibatch(n, ragged=ragged)):  # not order preserving
            pipeline.append(b)  # order preserving within pipeline                        
            if k < m-1:
                if prepad:
                    yield( (None, b) )  
            else:
                yield( (pipeline.pop(0), b) )  # yield deque-like (minibatch, shifted minibatch) tuples
        for p in pipeline:
            if postpad:
                yield( (p, None) )


    def chunks(self, sizes):
        """Partition the dataset into chunks of size given by the tuple in partitions, and give the dataset suffix if provided"""
        assert sum(sizes) == len(self)

        i = 0
        datasets = []
        for n in sizes:
            datasets.append(self.clone().index(self.index()[i:i+n]))
            i += n
        return datasets
        
    def partition(self, trainfraction=0.9, valfraction=0.1, testfraction=0, trainsuffix=':train', valsuffix=':val', testsuffix=':test'):
        """Partition the dataset into the requested (train,val,test) fractions.  

        Args:
            trainfraction [float]: fraction of dataset for training set
            valfraction [float]: fraction of dataset for validation set
            testfraction [float]: fraction of dataset for test set
            trainsuffix: If not None, append this string the to trainset ID
            valsuffix: If not None, append this string the to valset ID
            testsuffix: If not None, append this string the to testset ID        
        
        Returns:        
            (trainset, valset, testset) such that trainset is the first trainfraction of the dataset.  

        .. note:: This does not permute the dataset.  To randomize split, shuffle dataset first

        """
        assert trainfraction >=0 and trainfraction <= 1, "invalid training set fraction '%f'" % trainfraction
        assert valfraction >=0 and valfraction <= 1, "invalid validation set fraction '%f'" % valfraction
        assert testfraction >=0 and testfraction <= 1, "invalid test set fraction '%f'" % testfraction
        assert abs(trainfraction + valfraction + testfraction - 1) < 1E-6, "fractions must sum to one"
        
        idx = self.index()
        (testidx, validx, trainidx) = dividelist(idx, (testfraction, valfraction, trainfraction))
            
        trainset = self.clone().index(trainidx)
        if trainsuffix and trainset.id():
            trainset.id(trainset.id() + trainsuffix)
        
        valset = self.clone().index(validx)
        if valsuffix and valset.id():
            valset.id(valset.id() + valsuffix)
        
        testset = self.clone().index(testidx)
        if testsuffix and testset.id():
            testset.id(testset.id() + testsuffix)
                
        return (trainset,valset,testset) if testfraction!=0 else (trainset, valset)

    def split(self, size):
        """Split the dataset into two datasets, one of length size, the other of length len(self)-size"""
        assert isinstance(size, int) and size>=0 and size<len(self)
        return self.partition(size/len(self), (len(self)-size)/len(self), 0, '', '', '')

    def streaming_map(self, mapper, accepter=None, bufsize=1024):
        """Returns a generator that will apply the mapper and yield only those elements that return True from the accepter.  Performs the map in parallel if used in the vipy.globals.parallel context manager"""
        return self.__parallel_iter__(mapper=mapper, accepter=accepter, bufsize=bufsize)
        
    def map(self, f_map, strict=True, oneway=False, ordered=False):        
        """Parallel map.

        To perform this in parallel across four threads:

        ```python
        D = vipy.dataset.Dataset(...)
        with vipy.globals.parallel(4):
            D = D.map(lambda v: ...)
        ```

        Args:
            f_map: [lambda] The lambda function to apply in parallel to all elements in the dataset.  This must return a JSON serializable object (or set oneway=True)
            strict: [bool] If true, raise exception on distributed map failures, otherwise the map will return only those that succeeded
            oneway: [bool] If true, do not pass back results unless exception.  This is useful for distributed processing
        
        Returns:
            A `vipy.dataset.Dataset` containing the elements f_map(v).  This operation is order preserving if ordered=True.

        .. note:: 
            - This method uses dask distributed and `vipy.batch.Batch` operations
            - Due to chunking, all error handling is caught by this method.  Use `vipy.batch.Batch` to leverage dask distributed futures error handling.
            - Operations must be chunked and serialized because each dask task comes with overhead, and lots of small tasks violates best practices
            - Serialized results are deserialized by the client and returned a a new dataset
        """
        assert f_map is None or callable(f_map), "invalid map function"

        # Identity
        if f_map is None:
            return self        

        # Parallel map 
        elif vipy.globals.cf() is not None:
            # This will fail on multiprocessing if dataset contains a loader lambda, or any element in the dataset contains a loader.  Use distributed instead
            assert ordered == False, "not order preserving, use localmap()"
            return Dataset(tuple(vipy.parallel.map(f_map, self)), id=self.id()) 
                                              
        # Distributed map
        elif vipy.globals.dask() is not None:
            from vipy.batch import Batch   # requires pip install vipy[all]                
            f_serialize = lambda x: x
            f_deserialize = lambda x: x
            f_oneway = lambda x, oneway=oneway: x if not x[0] or not oneway else (x[0], None)
            f_catcher = lambda f, *args, **kwargs: catcher(f, *args, **kwargs)  # catch exceptions when executing lambda, return (True, result) or (False, exception)
            f = lambda x, f_serializer=f_serialize, f_deserializer=f_deserialize, f_map=f_map, f_catcher=f_catcher, f_oneway=f_oneway: f_serializer(f_oneway(f_catcher(f_map, f_deserializer(x))))  # with closure capture
            
            S = [f_serialize(v) for v in self]  # local load, preprocess and serialize
            B = Batch(chunklist(S, 128), strict=False, warnme=False, minscatter=128)
            S = B.map(lambda X,f=f: [f(x) for x in X]).result()  # distributed, chunked, with caught exceptions, may return empty list
            V = [f_deserialize(x) for s in S for x in s]  # Local deserialization and chunk flattening
            
            # Error handling
            (good, bad) = ([r for (b,r) in V if b], [r for (b,r) in V if not b])  # catcher returns (True, result) or (False, exception string)
            if len(bad)>0:
                log.warning('Exceptions in distributed processing:\n%s\n\n[vipy.dataset.Dataset.map]: %d/%d items failed' % (str(bad), len(bad), len(self)))
                if strict:
                    raise ValueError('exceptions in distributed processing')
            return Dataset(good, id=self.id()) if not oneway else None

        # Local map
        else:
            return self.localmap(f_map)
        
    def localmap(self, f):
        """A map performed without any parallel processing"""
        return Dataset([f(x) for x in self], id=self.id())  # triggers load into memory        

    def zip(self, iter):
        """Returns a new dataset constructed by applying the callable on elements from zip(self,iter)"""
        return Dataset(zip(self,iter))
    
    def sort(self, f):
        """Sort the dataset in-place using the sortkey lambda function f

        To perform a sort of the dataset using some property of the instance, such as the object category (e.g. for vipy.image.ImageCategory) 

        ```python
        dataset.sort(lambda im: im.category())
        ```
        """
        idx = self.index()
        return self.index( [idx[j] for (j,x) in sorted(zip(range(len(self)), self.tuple(f)), key=lambda x: x[1])] )

    
    @staticmethod
    def uniform_shuffler(D):
        """A uniform shuffle on the dataset elements.  Iterable access will be slow due to random access"""
        idx = D.index()
        random.shuffle(idx)
        return D.index(idx)

    @staticmethod
    def streaming_shuffler(D):
        """A uniform shuffle (approximation) on the dataset elements for iterable access only"""
        assert D._idx is None, "streaming only"
        
        try_import('datasets', 'datasets'); from datasets import Dataset as HuggingfaceDataset;
        
        if isinstance(D._ds, (list, tuple)):
            D._ds = list(D._ds)
            random.shuffle(D._ds)  # in-place shuffle objects
                
        elif isinstance(D._ds, HuggingfaceDataset):
            # Special case: Arrow backed dataset            
            D._ds = D._ds.to_iterable_dataset()  # no random access
            D._ds.shuffle()  # approximate shuffling for IterableDataset is much more efficient for __iter__
        else:
            raise ValueError('shuffle error')
        return D
    
    @staticmethod
    def identity_shuffler(D):
        """Shuffler that does nothing"""
        return D


class Paged(Dataset):
    """ Paged dataset.

    A paged dataset is a dataset of length N=M*P constructed from M archive files (the pages) each containing P elements (the pagesize).  
    The paged dataset must be constructed with tuples of (pagesize, filename).  
    The loader will fetch, load and cache the pages on demand using the loader, preserving the most recently used cachesize pages

    ```python
    D = vipy.dataset.Paged([(64, 'archive1.pkl'), (64, 'archive2.pkl')], lambda x,y: ivy.load(y))
    ```

    .. note :: Shuffling this dataset is biased.  Shuffling will be performed to mix the indexes, but not uniformly at random.  The goal is to preserve data locality to minimize cache misses.
    """
    
    def __init__(self, pagelist, loader, id=None, strict=True, index=None, cachesize=32):        
        super().__init__(dataset=pagelist,
                         id=id,
                         loader=loader).index(index if index else list(range(sum([p[0] for p in pagelist]))))

        assert callable(loader), "page loader required"
        assert not strict or len(set([x[0] for x in self._ds])) == 1  # pagesizes all the same 

        self._cachesize = cachesize
        self._pagecache = {}
        self._ds = list(self._ds)
        self._pagesize = self._ds[0][0]  # (pagesize, pklfile) tuples        

    def shuffle(self, shuffler=None):
        """Permute elements while preserve page locality to minimize cache misses"""
        shuffler = shuffler if shuffler is not None else functools.partial(Paged.chunk_shuffler, chunksize=int(1.5*self._pagesize))
        return shuffler(self)
        
    def __getitem__(self, k):
        if isinstance(k, (int, np.uint64)):
            assert abs(k) < len(self._idx), "invalid index"
            page = self._idx[int(k)] // self._pagesize
            if page not in self._pagecache:
                self._pagecache[page] = self._loader(*self._ds[page])  # load and cache new page
                if len(self._pagecache) > self._cachesize:
                    self._pagecache.pop(list(self._pagecache.keys())[0])  # remove oldest
            x = self._pagecache[page][int(k) % self._pagesize]
            return x
        elif isinstance(k, slice):
            return [self[i] for i in range(len(self))[k.start if k.start else 0:k.stop if k.stop else len(self):k.step if k.step else 1]]  # expensive
        else:
            raise ValueError('invalid index type "%s"' % type(k))            

    def flush(self):
        self._pagecache = {}
        return self

    @staticmethod
    def chunk_shuffler(D, chunker, chunksize=64):
        """Split dataset into len(D)/chunksize non-overlapping chunks with some common property returned by chunker, shuffle chunk order and shuffle within chunks.  

           - If chunksize=1 then this is equivalent to uniform_shuffler
           - chunker must be a callable of some property that is used to group into chunks
            
        """
        assert callable(chunker)
        return D.randomize().sort(chunker).index([i for I in shufflelist([shufflelist(I) for I in chunkgenbysize(D.index(), chunksize)]) for i in I])
    
    
class Union(Dataset):
    """vipy.dataset.Union() class
    
    Common class to manipulate groups of vipy.dataset.Dataset objects in parallel

    Usage:
    
        >>> cifar10 = vipy.dataset.registry('cifar10')
        >>> mnist = vipy.dataset.registry('mnist')
        >>> dataset = vipy.dataset.Union(mnist, cifar10)
        >>> dataset = mnist | cifar10

    Args:
        Datasets 
    """

    __slots__ = ('_id', '_ds', '_idx', '_loader', '_type')    
    def __init__(self, *args, **kwargs):
        assert all(isinstance(d, (Dataset, )) for d in args), "invalid datasets"
        
        datasets = [d for d in args]  # order preserving
        assert all([isinstance(d, Dataset) for d in datasets]), "Invalid datasets '%s'" % str([type(d) for d in datasets])

        datasets = [j for i in datasets for j in (i.datasets() if isinstance(i, Union) else (i,))]  # flatten unions        
        self._ds = datasets
        self._idx = None
        self._id = kwargs['id'] if 'id' in kwargs else None

        self._loader = None  # individual datasets have loaders
        self._type = None

    def is_streaming(self):
        return self._idx is None and all(d.is_streaming() for d in self.datasets())

    def __len__(self):
        return sum(d.__len__() for d in self.datasets()) if self._idx is None else len(self._idx)
    
    def __iter__(self, mapper=None, accepter=None):
        if self.is_streaming():
            k = -1
            iter = [d.__iter__(mapper, accepter) for d in self.datasets()]  # round-robin

            for m in range(len(self.datasets())):
                try:
                    while True:
                        k = (k + 1) % len(iter)                                    
                        yield next(iter[k])  # assumes ordered
                except StopIteration:
                    iter.pop(k)
                    k -= 1
            
        else:
            self.index()  # force random access                    
            for (i,j) in self._idx:
                x = self._ds[i][j]  # random access (slower)                
                x = mapper(x) if mapper is not None else x
                if accepter is None or accepter(x):
                    yield x                  

    def __getitem__(self, k):
        self.index()  # force random access        
        if isinstance(k, (int, np.uint64)):
            assert abs(k) < len(self._idx), "invalid index"
            (i,j) = self._idx[int(k)]            
            return self._ds[i][j]
        elif isinstance(k, slice):
            return [self._ds[i][j] for (i,j) in self._idx[k.start:k.stop:k.step]]
        else:
            raise ValueError('invalid index type "%s"' % type(k))

    def __repr__(self):
        fields = ['id=%s' % self.id(truncated=True, maxlen=64)] if self.id() else []
        fields += ['len=%d' % len(self)]
        fields += ['union=%s' % str(tuple([d.id(truncated=True, maxlen=80) for d in self._ds]))]
        return str('<vipy.dataset.Dataset: %s>' % (', '.join(fields)))

    def index(self, index=None, strict=False):
        """Update the index, useful for filtering of large datasets"""
        if index is not None:
            self._idx = index
            return self
        if self._idx is None:
            # Index on-demand: zipped (dataset index, element index) tuples, in round-robin dataset order [(0,0),(1,0),...,(0,n),(1,n),...]            
            lengths = [len(d) for d in self.datasets()]            
            self._idx = [c for r in [[(i,j) for i in range(len(self.datasets()))] for j in range(max(lengths))] for c in r if c[1]<lengths[c[0]]]
        return self._idx
    
    def clone(self, deep=False):
        """Return a copy of the dataset object"""
        D = super().clone(deep=deep)
        D._ds =  [d.clone(deep=deep) for d in D._ds]
        return D
    
    def datasets(self):
        """Return the dataset union elements, useful for generating unions of unions"""
        return list(self._ds)

    def shuffle(self, shuffler=None):
        """Permute elements in this dataset uniformly at random in place using the best shuffler for the dataset structure"""
        shuffler = shuffler if shuffler is not None else (Union.streaming_shuffler if self.is_streaming() else Dataset.uniform_shuffler)
        return shuffler(self)
    
    @staticmethod
    def streaming_shuffler(D):
        """A uniform shuffle (approximation) on the dataset elements for iterable access only"""
        assert D._idx is None, "iterable dataset only"
        D._ds = [Dataset.streaming_shuffler(d) for d in D._ds]  # shuffle dataset shards
        random.shuffle(D._ds)  # shuffle union order
        return D
    

def registry(name=None, datadir=None, freeze=True, clean=False, download=False, split='train'):
    """Common entry point for loading datasets by name.

    Usage:
    
        >>> trainset = vipy.dataset.registry('cifar10', split='train')             # return a training split
        >>> valset = vipy.dataset.registry('cifar10:val', datadir='/tmp/cifar10')  # download to a custom location
        >>> datasets = vipy.dataset.registry(('cifar10:train','cifar100:train'))   # return a union
        >>> vipy.dataset.registry()                                                # print allowable datasets

    Args:
       name [str]: The string name for the dataset.  If tuple, return a `vipy.dataset.Union`.  If None, return the list of registered datasets.  Append name:train, name:val, name:test to output the requested split, or use the split keyword.
       datadir [str]: A path to a directory to store data.  Defaults to environment variable VIPY_DATASET_REGISTRY_HOME (then VIPY_CACHE if not found).  Also uses HF_HOME for huggingface datasets.  Datasets will be stored in datadir/name
       freeze [bool]:  If true, disable reference cycle counting for the loaded object (which will never contain cycles anyway) 
       clean [bool]: If true, force a redownload of the dataset to correct for partial download errors
       download [bool]: If true, force a redownload of the dataset to correct for partial download errors
       split [str]: return 'train', 'val' or 'test' split.  If None, return (trainset, valset, testset) tuple

    Datasets:
       'mnist','cifar10','cifar100','caltech101','caltech256','oxford_pets','sun397', 'food101','stanford_dogs',
       'flickr30k','oxford_fgvc_aircraft','oxford_flowers_102','eurosat','d2d','ethzshapes','coil100','kthactions',
       'yfcc100m','yfcc100m_url','tiny_imagenet','coyo300m','coyo700m','pascal_voc_2007','coco_2014', 'ava',
       'activitynet', 'open_images_v7', 'imagenet', 'imagenet21k', 'visualgenome' ,'widerface','meva_kf1',
       'objectnet','lfw','inaturalist_2021','kinetics','hmdb','places365','ucf101','lvis','kitti',
       'imagenet_localization','laion2b','datacomp_1b','imagenet2014_det','imagenet_faces','youtubeBB',
       'pip_370k','pip_175k','cap','cap_pad','cap_detection','tiny_virat'

    Returns:
       (trainset, valset, testset) tuple where each is a `vipy.dataset.Dataset` or None, or a single split if name has a ":SPLIT" suffix or split kwarg provided
    """
    
    import vipy.data

    datasets = ('mnist','cifar10','cifar100','caltech101','caltech256','oxford_pets','sun397', 'stanford_dogs','coil100',
                'flickr30k','oxford_fgvc_aircraft','oxford_flowers_102', 'food101', 'eurosat','d2d','ethzshapes','kthactions',
                'yfcc100m','yfcc100m_url','tiny_imagenet','coyo300m','coyo700m','pascal_voc_2007','coco_2014', 'ava',
                'activitynet','open_images_v7','imagenet','imagenet21k','visualgenome','widerface', 'youtubeBB',
                'objectnet','lfw','inaturalist_2021','kinetics','hmdb','places365','ucf101','kitti','meva_kf1',
                'lvis','imagenet_localization','laion2b','datacomp_1b','imagenet2014_det','imagenet_faces',
                'pip_175k','pip_370k','cap','cap_pad','cap_detection','tiny_virat')  # Add to docstring too...
    
    if name is None:
        return tuple(sorted(datasets))
    if isinstance(name, (tuple, list)):
        assert all(n.startswith(datasets) for n in name)
        assert split is not None or all(':' in n for n in name)
        return Union(*(registry(n, datadir=datadir, freeze=freeze, clean=clean, download=download, split=split) for n in name))    
    
    (name, split) = name.split(':',1) if name.count(':')>0 else (name, split)
    if name not in datasets:
        raise ValueError('unknown dataset "%s" - choose from "%s"' % (name, ', '.join(sorted(datasets))))
    if split not in [None, 'train', 'test', 'val']:
        raise ValueError('unknown split "%s" - choose from "%s"' % (split, ', '.join([str(None), 'train', 'test', 'val'])))

    datadir = remkdir(datadir if datadir is not None else (env('VIPY_DATASET_REGISTRY_HOME') if 'VIPY_DATASET_REGISTRY_HOME' in env() else cache()))
    namedir = Path(datadir)/name    
    if (clean or download) and name in datasets and os.path.exists(namedir):
        log.info('Removing cached dataset "%s"' % namedir)
        shutil.rmtree(namedir)  # delete cached subtree to force redownload ...
        
    if freeze:
        gc.disable()
        
    (trainset, valset, testset) = (None, None, None)    
    if name == 'mnist':
        (trainset, testset) = vipy.data.hf.mnist()        
    elif name == 'cifar10':
        (trainset, testset) = vipy.data.hf.cifar10()        
    elif name == 'cifar100':
        (trainset, testset) = vipy.data.hf.cifar100()        
    elif name == 'caltech101':
        trainset = vipy.data.caltech101.Caltech101(namedir)        
    elif name == 'caltech256':
        trainset = vipy.data.caltech256.Caltech256(namedir)
    elif name == 'oxford_pets':
        (trainset, testset) = vipy.data.hf.oxford_pets()
    elif name == 'sun397':
        (trainset, valset, testset) = vipy.data.hf.sun397()
    elif name == 'stanford_dogs':
        trainset = vipy.data.stanford_dogs.StanfordDogs(namedir)
    elif name == 'food101':
        trainset = vipy.data.food101.Food101(namedir)
    elif name == 'eurosat':
        trainset = vipy.data.eurosat.EuroSAT(namedir)
    elif name == 'd2d':
        trainset = vipy.data.d2d.D2D(namedir)
    elif name == 'coil100':
        trainset = vipy.data.coil100.COIL100(namedir)
    elif name == 'kthactions':
        (trainset, testset) = vipy.data.kthactions.KTHActions(namedir).split()
    elif name == 'ethzshapes':
        trainset = vipy.data.ethzshapes.ETHZShapes(namedir)        
    elif name == 'flickr30k':
        trainset = vipy.data.hf.flickr30k()
    elif name == 'oxford_fgvc_aircraft':
        trainset = vipy.data.hf.oxford_fgvc_aircraft()
    elif name == 'oxford_flowers_102':
        trainset = vipy.data.oxford_flowers_102.Flowers102(namedir)
    elif name == 'yfcc100m':
        (trainset, _, valset, _) = vipy.data.hf.yfcc100m()  
    elif name == 'yfcc100m_url':
        (_, trainset, _, valset) = vipy.data.hf.yfcc100m()  
    elif name == 'tiny_imagenet':
        (trainset, valset) = vipy.data.hf.tiny_imagenet()
    elif name == 'coyo300m':
        trainset = vipy.data.hf.coyo300m()
    elif name == 'coyo700m':
        trainset = vipy.data.hf.coyo700m()
    elif name == 'datacomp_1b':
        trainset = vipy.data.hf.datacomp_1b()
    elif name == 'pascal_voc_2007':
        (trainset, valset, testset) = vipy.data.hf.pascal_voc_2007()
    elif name == 'coco_2014':
        trainset = vipy.data.coco.COCO_2014(namedir)
    elif name == 'ava':
        ava = vipy.data.ava.AVA(namedir)
        (trainset, valset) = (ava.trainset(), ava.valset())
    elif name == 'activitynet':
        activitynet = vipy.data.activitynet.ActivityNet(namedir)  # ActivityNet 200
        (trainset, valset, testset) = (activitynet.trainset(), activitynet.valset(), activitynet.testset())
    elif name == 'open_images_v7':
        trainset = vipy.data.openimages.open_images_v7(namedir)
    elif name == 'imagenet':
        imagenet = vipy.data.imagenet.Imagenet2012(namedir)
        (trainset, valset) = (imagenet.classification_trainset(), imagenet.classification_valset())
    elif name == 'imagenet_faces':
        trainset = vipy.data.imagenet.Imagenet2012(Path(datadir)/'imagenet').faces()
    elif name == 'imagenet21k':
        trainset = vipy.data.imagenet.Imagenet21K(namedir)
    elif name == 'visualgenome':
        trainset = vipy.data.visualgenome.VisualGenome(namedir)  # visualgenome-1.4
    elif name == 'widerface':
        trainset = vipy.data.widerface.WiderFace(namedir, split='train')
        valset = vipy.data.widerface.WiderFace(namedir, split='val')
        testset = vipy.data.widerface.WiderFace(namedir, split='test')                                      
    elif name == 'objectnet':
        assert split is None or split == 'test', "objectnet is a test set"
        testset = vipy.data.objectnet.Objectnet(namedir)
    elif name == 'lfw':
        trainset = vipy.data.lfw.LFW(namedir)
    elif name == 'inaturalist_2021':
        dataset = vipy.data.inaturalist.iNaturalist2021(namedir)
        (trainset, valset) = (dataset.trainset(), dataset.valset())
    elif name == 'kinetics':
        dataset = vipy.data.kinetics.Kinetics700(namedir)  # Kinetics700
        (trainset, valset, testset) = (dataset.trainset(), dataset.valset(), dataset.testset())
    elif name == 'hmdb':
        trainset = vipy.dataset.Dataset(vipy.data.hmdb.HMDB(namedir).dataset(), id='hmdb')
    elif name == 'places365':
        places = vipy.data.places.Places365(namedir)
        (trainset, valset) = (places.trainset(), places.valset())
    elif name == 'ucf101':
        trainset = vipy.data.ucf101.UCF101(namedir)
    elif name == 'kitti':
        trainset = vipy.data.kitti.KITTI(namedir, split='train')
        testset = vipy.data.kitti.KITTI(namedir, split='test')                                      
    elif name == 'lvis':
        lvis = vipy.data.lvis.LVIS(namedir)
        (trainset, valset) = (lvis.trainset(), lvis.valset())
    elif name == 'imagenet_localization':
        trainset = vipy.data.imagenet.Imagenet2012(Path(datadir)/'imagenet').localization_trainset()
    elif name == 'imagenet2014_det':
        imagenet2014_det = vipy.data.imagenet.Imagenet2014_DET(namedir)
        (trainset, valset, testset) = (imagenet2014_det.trainset(), imagenet2014_det.valset(), imagenet2014_det.testset())
    elif name == 'laion2b':
        trainset = vipy.data.hf.laion2b()
    elif name == 'youtubeBB':
        trainset = vipy.data.youtubeBB.YoutubeBB(namedir)
    elif name == 'meva_kf1':
        trainset = vipy.data.meva.KF1(namedir).dataset()  # consider using "with vipy.globals.multiprocessing(pct=0.5):"
    elif name == 'pip_175k':
        trainset = vipy.data.pip.PIP_175k(namedir)
    elif name == 'pip_370k':
        trainset = vipy.data.pip.PIP_370k_stabilized(namedir)
    elif name == 'cap':
        trainset = vipy.data.cap.CAP_classification_clip(namedir)
    elif name == 'cap_pad':
        trainset = vipy.data.cap.CAP_classification_pad(namedir)        
    elif name == 'cap_detection':
        trainset = vipy.data.cap.CAP_detection(namedir)
    elif name == 'tiny_virat':
        dataset = vipy.data.tiny_virat.TinyVIRAT(namedir)
        (trainset, valset, testset) = (dataset.trainset(), dataset.valset(), dataset.testset())
    else:
        raise ValueError('unknown dataset "%s" - choose from "%s"' % (name, ', '.join(sorted(datasets))))
    
    if freeze:
        gc.enable()
        gc.collect()
        gc.freeze()  # python-3.7

    if split == 'train':
        return trainset
    elif split == 'val':
        return valset
    elif split == 'test':
        return testset
    else:
        return (trainset, valset, testset)



import numpy as np
from vipy.geometry import BoundingBox
from vipy.util import isstring, tolist, shortuuid
import uuid
import copy
from vipy.object import Track

try:
    import ujson as json  # faster
except ImportError:
    import json


class Activity(object):
    """vipy.object.Activity class
    
    An activity is a grouping of one or more tracks involved in an activity within a given startframe and endframe.
    The activity occurs at a given (startframe, endframe), where these frame indexes are extracted at the provided framerate.
    All objects are passed by reference with a globally unique track ID, for the tracks involved with the activity.  This 
    is done since tracks can exist after an activity completes, and that tracks should update the spatial transformation of boxes.
        
    Valid constructors

    ```python 
    t = vipy.object.Track(category='Person')
    a = vipy.object.Activity(startframe=0, endframe=10, category='Walking', tracks=t.id())
    a = vipy.object.Activity(startframe=0, endframe=10, category='Walking', tracks=t)    
    ```

    Note.. shortlabel is kepy for backwards compatibility and will be deprecated
    
    """
    __slots__ = ['_id', '_startframe', '_endframe', '_framerate', '_label', '_trackid', '_actorid', 'attributes']
    def __init__(self, startframe, endframe, framerate=30, label=None, category=None, tracks=None, attributes=None, actorid=None, confidence=None, id=None, shortlabel=None):
        assert not (label is not None and category is not None), "Activity() Constructor requires either label or category kwargs, not both"
        assert startframe <= endframe, "Start frame must be less than or equal to end frame"
        assert framerate is not None, "initial framerate is required for framerate conversion"
        if tracks:
            tracks = [v for (k,v) in tracks.items()] if isinstance(tracks, dict) else tracks  # backwards compatible, track dictionary input
            assert (all([isstring(t) for t in tracks]) or all([isinstance(t, Track) for t in tracks])), "Invalid track input"
            if all([isinstance(t, Track) for t in tracks]):
                assert any([any([t.during(f) for f in range(startframe, endframe)]) for t in tracks]), "At least one track must be be present in at least one frame when this activity occurs"
                tracks = [t.id() for t in tracks]  
        trackid = set(tracks) if tracks is not None else (set([actorid]) if actorid is not None else set([]))  # only store IDs, not tracks
        
        if tracks is not None and actorid is not None and actorid not in trackid:
            trackid.add(actorid)

        self._id = shortuuid() if id is None else id  # use provided
        self._startframe = int(startframe)
        self._endframe = int(endframe)
        self._framerate = float(framerate)
        self._label = category if category is not None else label        
        self._trackid = trackid
        self._actorid = actorid

        self.attributes = attributes.copy() if attributes is not None else {}  # shallow copy
        if confidence is not None:
            self.attributes['confidence'] = float(confidence)

    def hasattribute(self, k):
        return k in self.attributes

    def confidence(self, c=None):
        if c is not None:
            self.attributes['confidence'] = float(c)
            return self
        else:
            return self.attributes['confidence'] if 'confidence' in self.attributes else None
    
    @classmethod
    def from_json(obj, s):
        d = json.loads(s) if not isinstance(s, dict) else s
        d = {k.lstrip('_'):v for (k,v) in d.items()}  # prettyjson (remove "_" prefix to attributes)                                        
        return obj(startframe=int(d['startframe']),
                   endframe=int(d['endframe']),
                   framerate=d['framerate'] if 'framerate' in d else None,
                   category=d['label'] if 'label' in d else None,
                   tracks=d['trackid'] if 'trackid' in d else None,
                   attributes=d['attributes'] if 'attributes' in d else None,
                   actorid=d['actorid'] if 'actorid' in d else None,
                   id=d['id'] if 'id' in d else None)
                
    def __len__(self):
        """Return activity length in frames, or zero if degenerate"""
        return max(0, self.endframe() - self.startframe())

    def duration(self, s=None, centered=False):
        """The length of the activity in seconds.

        Args:
            s: [float] The number of seconds for this activity, starting at the startframe
            centered: [bool] If true, then set the duration centered on the middle frame

        Returns:
            The duration in seconds of this activity object (if s=None)
            This activity object with the requested duration (if s!=None)
        """
        assert self.framerate() is not None, "Framerate must be set in constructor"
        if s is None:
            return len(self) / float(self.framerate())
        else:
            return (self.endframe(self.startframe() + int(round(s*self.framerate()))) if not centered else 
                    self.truncate(startframe=self.middleframe()-int(np.ceil(s*self.framerate())//2), endframe=self.middleframe()+int(np.ceil(s*self.framerate())//2)))
        
    def __repr__(self):
        return str('<vipy.activity.Activity: category="%s", frames=(%d,%d), tracks=%s%s>' % (self.category(), self.startframe(), self.endframe(), len(self.trackids()), '' if self.confidence() is None else ', confidence=%1.2f' % self.confidence()))

    def dict(self):
        """Return a python dictionary containing the relevant serialized attributes suitable for JSON encoding"""
        return self.json(s=None, encode=False)

    def __json__(self):
        """Serialization method for json package"""
        return self.json(encode=True)
    
    def json(self, encode=True):
        d = {k.lstrip('_'):getattr(self, k) for k in Activity.__slots__ if getattr(self, k) is not None}  # prettyjson (remove "_" prefix to attributes)          
        d = {k:v if k != 'trackid' else tuple(v) for (k,v) in d.items()}  # sets are non-serializable
        return json.dumps(d) if encode else d
    
    def actorid(self, actorid=None):
        if actorid is None:
            return self._actorid
        else:
            self._actorid = actorid
            if actorid not in self._trackid:
                self._trackid.add(actorid) if isinstance(self._trackid, set) else self._trackid.append(actorid)                    
            return self

    def startframe(self, f=None):
        if f is None:
            return self._startframe
        else:
            self._startframe = f
            return self
    
    def endframe(self, f=None):
        if f is None:
            return self._endframe
        else:
            self._endframe = int(f)
            return self

    def middleframe(self):
        """Return the middle frame number of the activity"""
        return int(np.round((self._endframe - self._startframe) / 2.0)) + self._startframe

    def _set_framerate(self, fps):
        """Override framerate conversion and just set the framerate attribute.  This should really be set only in constructor.  Use with caution!"""
        self._framerate = float(fps)
        return self

    def framerate(self, fps=None, speed=None):
        """Resample (startframe, endframe) from known original framerate set by constructor to be new framerate fps"""        
        if fps is None and speed is None:
            return self._framerate
        else:
            assert fps is not None or speed is not None, "Invalid input"
            assert not (fps is not None and speed is not None), "Invalid input"
            assert speed is None or speed > 0, "Invalid speed, must specify speed multiplier s=1, s=2 for 2x faster, s=0.5 for half slower"                    
            assert fps is None or fps > 0, "Invalid framerate"
            assert self._framerate is not None, "Framerate conversion requires that the framerate is known for current activities.  This must be provided to the vipy.object.Activity() constructor."

            fps = float(fps) if fps is not None else (1.0/speed)*self._framerate            
            (self._startframe, self._endframe) = [int(np.round(f*(fps/float(self._framerate)))) for f in (self._startframe, self._endframe)]
            self._framerate = float(fps)
            return self
    
    def category(self, label=None):
        """Change the label to the new label """
        if label is not None:
            self._label = label
            return self
        else:
            return self._label

    def categoryif(self, ifcategory, tocategory=None):
        """If the current category is equal to ifcategory, then change it to newcategory.

        Args:
            
            ifcategory [dict, str]: May be a dictionary {ifcategory:tocategory}, or just an ifcategory
            tocategory [str]:  the target category 

        Returns:
        
            this object with the category changed.

        .. note:: This is useful for converting synonyms such as self.categoryif('person_sits', 'person_sitting')
        """
        assert (isinstance(ifcategory, dict) and tocategory is None) or tocategory is not None

        if isinstance(ifcategory, dict):
            for (k,v) in ifcategory.items():
                self.categoryif(k, v)
        elif self.category() == ifcategory:
            self.category(tocategory)
        return self

    def label(self, label=None):
        """Alias for category"""
        return self.category(label)

    def add(self, track):
        """Add the track id for the track to this activity, so that if the track is changed externally it is reflected here"""
        assert isinstance(track, Track), "Invalid input - must be vipy.object.Track"
        assert self.during_interval(track.startframe(), track.endframe()), "The track must be present during the activity"
        if track.id() not in self._trackid:
            self._trackid = list(self._trackid)
            self._trackid.append(track.id())
        return self

    def addid(self, trackid):
        """Add the track id for the track to this activity, so that if the track is changed externally it is reflected here"""
        if trackid not in self._trackid:
            self._trackid = list(self._trackid)
            self._trackid.append(trackid)
        return self
    
    def tracks(self):
        """alias for trackids"""
        return self.trackids()

    def cleartracks(self):
        """Remove all track IDs from this activity"""
        self._trackid = []
        return self

    def trackids(self):
        """Return a set of track IDs associated with this activity"""
        return set(self._trackid)

    def hasoverlap(self, other, threshold=0):
        """Return true if the temporal_iou is greater than the provided threshold between self and other Track or other Activity"""
        assert isinstance(other, Activity) or isinstance(other, Track), "Invalid input"
        assert threshold >= 0 and threshold <= 1, "Invalid temporal IOU threshold"
        return (((min(self._endframe, other.endframe()) - max(self._startframe, other.startframe())) > 0) if threshold == 0 else
                self.temporal_iou(other) > threshold)
        
    def isneighbor(self, other, framegate=10):
        return self.temporal_iou(other.clone().temporalpad(framegate)) > 0 

    def hastrack(self, track):
        """Is the track part of the activity?"""
        assert isstring(track) or isinstance(track, Track), "Invalid input - Must be a vipy.object.Track().id() or vipy.object.Track()"
        trackid = track.id() if isinstance(track, Track) else track
        return trackid in self._trackid

    def hastrackoverlap(self, track):
        """is the activity occurring during the interval when the track is occurring and is this track assigned to the activity?"""
        assert isinstance(track, Track)
        return self.hastrack(track) and self.temporal_iou(track) > 0

    def append(self, newtrack):
        """Append newtrack to this activity and set as actorid()"""
        assert isinstance(newtrack, Track), "Invalid input - must be vipy.object.Track"
        self._trackid.add(newtrack.id()) if isinstance(self._trackid, set) else self._trackid.append(newtrack.id())
        self.actorid(newtrack.id())
        return self

    def trackfilter(self, f):
        """Remove all tracks such that the lambda function f(trackid) resolves to False"""
        assert callable(f)
        self._trackid = [tid for tid in self._trackid if f(tid)]
        if self.actorid() not in self._trackid:
            self._actorid = None
        return self

    def replace(self, oldtrack, newtrack):
        """Replace oldtrack with newtrack if present in self._tracks.  Pass in a trackdict to share reference to track, so that track owner can modify the track and this object observes the change"""
        assert isinstance(oldtrack, Track) and isinstance(newtrack, Track), "Invalid input - must be vipy.object.Track"
        if self.hastrack(oldtrack):
            self._trackid.discard(oldtrack.id())
            self._trackid.add(newtrack.id())
            if self.actorid() == oldtrack.id():
                self.actorid(newtrack.id())
        return self    

    def replaceid(self, oldtrackid, newtrackid):
        """Replace oldtrack with newtrack if present in self._tracks.  Pass in a trackdict to share reference to track, so that track owner can modify the track and this object observes the change"""
        if self.hastrack(oldtrackid):
            ti = set(self._trackid)
            ti.discard(oldtrackid)
            ti.add(newtrackid)
            if self.actorid() == oldtrackid:
                self.actorid(newtrackid)
            self._trackid = list(ti)
        return self    
    
    def during(self, frame):
        """Is frame during the time interval (startframe, endframe) inclusive?"""
        return int(frame) >= self._startframe and int(frame) <= self._endframe

    def during_interval(self, startframe, endframe, inclusive=False):
        """Is the activity occurring for any frames within the interval [startframe, endframe) (non-inclusive of endframe)?"""
        ef = endframe+(0 if not inclusive else 1)
        return (startframe >= self._startframe and startframe <= self._endframe) or (ef >= self._startframe and ef <= self._endframe) or (startframe <= self._startframe and ef >= self._endframe)

    
    def union(self, other, confweight=0.5, maxconf=False):
        """Compute the union of the new activity other to this activity by updating the start and end times and computing the mean confidence.
        
           -Note: other must have the same category and track IDs as self
           -confweight [0,1]:  the convex combinatiopn weight applied to the new activity 
        """
        assert isinstance(other, Activity), "Invalid input"
        assert self._actorid == other._actorid, "Actor ID must be the same"
        assert self._label == other._label, "Assigned activity is a different category"
        assert self._framerate == other._framerate, "Invalid input"
        assert confweight >= 0 and confweight <= 1, "Confidence weight must be [0,1]"

        self.startframe(min(other._startframe, self._startframe))
        self.endframe(max(other._endframe, self._endframe))
        if other.confidence() is not None and self.confidence() is not None:
            self.confidence(float((1.0-confweight)*self.confidence() + confweight*other.confidence()) if not maxconf else float(max(self.confidence(), other.confidence())))  # running mean confidence or max
        return self
        
    def temporal_iou(self, other):
        """Return the temporal intersection over union of two activities or this activity and a track"""
        assert isinstance(other, Activity) or isinstance(other, Track), "Invalid input - must be vipy.activity.Activity or vipy.object.Track"
        assert self._framerate == other._framerate, "invalid input - framerate must match"
        (sf, ef) = (other._startframe, other._endframe) if isinstance(other, Activity) else (other.startframe(), other.endframe())  # attribute access is faster than methods
        
        t_start = min(self._startframe, sf)
        t_end = max(self._endframe, ef)
        t_union = float(t_end - t_start)
        
        t_start = max(self._startframe, sf)
        t_end = min(self._endframe, ef)
        t_intersection = float(t_end - t_start)
        
        return (t_intersection / t_union) if t_intersection > 0 else 0
    
    def offset(self, dt):
        dt = dt if isinstance(dt, int) else int(np.round(dt*self.framerate()))
        self._startframe = int(self._startframe + dt)
        self._endframe = int(self._endframe + dt)
        return self

    def truncate(self, start=None, end=None):
        """Truncate the activity so that it is between start and end"""
        startframe = int(np.round(start*self.framerate())) if isinstance(start, float) else start
        endframe = int(np.round(end*self.framerate())) if isinstance(end, float) else end        
        
        self._startframe = self._startframe if startframe is None else max(self._startframe, startframe)
        self._endframe = self._endframe if endframe is None else min(self._endframe, endframe)
        self._endframe = self._endframe if self._endframe > self._startframe else self._startframe  # degenerate truncation
        return self
            
    def id(self, newid=None):
        if newid is None:
            return self._id
        else:
            self._id = newid
            return self

    def clone(self, rekey=False):
        #a = copy.deepcopy(self)
        a = Activity.from_json(self.json(encode=False))
        if rekey:
            a.id(newid=shortuuid())
        return a
    
    def temporalpad(self, df):
        """Add a temporal pad of df=(before frames, after frames) or df=pad frames to the start and end of the activity.  The padded start frame may be negative."""
        df = (df, df) if not isinstance(df, tuple) else df
        self._startframe -= int(df[0])
        self._endframe += int(df[1])
        return self  

    def padto(self, t):
        """Add a symmetric temporal pad so that the activity is at least t seconds long"""
        return self.temporalpad(int(np.ceil(self.framerate()*((t - self.duration())/2.0)))) if t > self.duration() else self

    def disjoint(self, other, strict=False):
        """Enforce disjoint activities with other by shifting the endframe or startframe of self to not overlap if they share the same tracks.
           Other may be an Activity() or list of Activity()
           if strict=True, then throw an exception if other or self is fully contained with the other, resulting in degenerate activity after disjoint
        """
        for o in tolist(other):
            assert isinstance(o, Activity), "Invalid input - must be vipy.activity.Activity() or list of activities"       
            if strict:
                assert not (o.during(self.startframe()) and o.during(self.endframe())), "Self cannot fully overlap other"
                assert not (self.during(o.startframe()) and self.during(o.endframe())), "Other cannot fully overlap self"
            if o.trackids() == self.trackids() and o.during(self.endframe()):
                self.endframe(o.startframe()-1)
            if o.trackids() == self.trackids() and o.during(self.startframe()):
                self.startframe(o.endframe()+1)
        return self  # may be zero length now

    def temporal_distance(self, other):
        """Return the temporal distance in frames between self and other which is the minimum frame difference between the end of one to the start of the other, or zero if they overlap"""
        assert isinstance(other, Activity), "Invalid input - must be vipy.activity.Activity()"
        return (max(self.startframe(), other.startframe()) - min(self.endframe(), other.endframe())) if self.temporal_iou(other) == 0 else 0

    def within(self, startframe, endframe):
        """Is the activity within the frame rate (startframe, endframe)?"""
        return self.startframe() >= startframe and self.endframe() <= endframe
    



import os
from vipy.util import filetail, remkdir, readjson, groupbyasdict
import vipy.downloader
from vipy.video import VideoCategory, Video
import numpy as np


# http://activity-net.org/download.html
#URL = 'http://ec2-52-25-205-214.us-west-2.compute.amazonaws.com/files/activity_net.v1-3.min.json'
URL = 'https://github.com/activitynet/ActivityNet/raw/refs/heads/master/Evaluation/data/activity_net.v1-3.min.json'


class ActivityNet():
    def __init__(self, datadir=vipy.util.tocache('activitynet'), redownload=False):
        """Activitynet, provide a datadir='/path/to/store/activitynet' """
        self._url = URL
        self.datadir = remkdir(datadir)
        if redownload or not self._isdownloaded():
            self.download()
        
    def __repr__(self):
        return str('<vipy.data.activitynet: "%s">' % self.datadir)

    def download(self):
        vipy.downloader.download(URL, os.path.join(self.datadir, filetail(URL)))
        return self

    def _isdownloaded(self):
        return os.path.exists(os.path.join(self.datadir, 'activity_net.v1-3.min.json'))
    
    def _dataset(self, subset):
        assert self._isdownloaded(), "Dataset not downloaded.  download() first or manually download '%s' into '%s'" % (self._url, self.datadir)        
        jsonfile = os.path.join(self.datadir, filetail(URL))
        json = readjson(jsonfile)

        return [(v['url'],
                 os.path.join(self.datadir, youtubeid),
                 a['label'],
                 float(a['segment'][0]),
                 float(a['segment'][1]))
                for (youtubeid, v) in json['database'].items()
                for a in v['annotations']
                if v['subset'] == subset]

    def trainset(self):
        loader = lambda x: VideoCategory(url=x[0], filename=x[1], category=x[2], startsec=x[3], endsec=x[4], framerate=None)
        return vipy.dataset.Dataset(self._dataset('training'), id='activitynet:train', loader=loader)

    def testset(self):
        """ActivityNet test set does not include any annotations"""
        assert self._isdownloaded(), "Dataset not downloaded.  download() first or manually download '%s' into '%s'" % (self._url, self.datadir)        
        json = readjson(os.path.join(self.datadir, filetail(URL)))
        loader = lambda x: Video(url=x[0], filename=x[1], framerate=None)
        return vipy.dataset.Dataset([(v['url'], os.path.join(self.datadir, youtubeid)) for (youtubeid, v) in json['database'].items() if v['subset'] == 'testing'], id='activitynet:test', loader=loader)

    def valset(self):
        loader = lambda x: VideoCategory(url=x[0], filename=x[1], category=x[2], startsec=x[3], endsec=x[4], framerate=None)
        return vipy.dataset.Dataset(self._dataset('validation'), id='activitynet:val', loader=loader)
        
    def categories(self):
        return set([v.category() for v in self.trainset()])

    def analysis(self):
        C = self.categories()
        d_category_to_trainsize = {k:len(v) for (k,v) in groupbyasdict(self.trainset(), lambda x: x.category()).items()}

        top10 = sorted([(k,v) for (k,v) in d_category_to_trainsize.items()], key=lambda x: x[1])[-10:]
        print('top10 categories by number of instances in training set:')
        print(top10)

        bottom10 = sorted([(k,v) for (k,v) in d_category_to_trainsize.items()], key=lambda x: x[1])[0:10]
        print('bottom-10 categories by number of instances in training set:')
        print(bottom10)
        
        return d_category_to_trainsize



import os
from vipy.util import remkdir, tocache, readcsv, filetail, groupbyasdict, isinstalled
import vipy.downloader
from vipy.dataset import Dataset
from vipy.video import Scene
from vipy.object import Track, Detection


URLS = ['https://research.google.com/youtube-bb/yt_bb_classification_train.csv.gz',
        'https://research.google.com/youtube-bb/yt_bb_detection_train.csv.gz']


class YoutubeBB(Dataset):
    """https://research.google.com/youtube-bb/download.html

    Usage:

    To show the clips centered on a track at 30Hz
    >>> dataset = vipy.data.youtubeBB.YoutubeBB()
    >>> v = dataset.takeone().download()
    >>> for t in v.trackclip():
    >>>     t.show()

    To extract the first annotated keyframes at 1Hz:
    >>> imlist = [t.framerate(1).frame(0) for t in dataset.takeone().trackclip()]
    >>> imlist[0].show()

    To extract all annotated keyframes at 1Hz for each tracked object:
    >>> framelist = [[im for im in t.framerate(1).load()] for t in dataset.takeone().trackclip()]
    >>> vipy.visualize.montage([im.centersquare().annotate() for frames in framelist for im in frames]).show()

    The boxes are available as:
    >>> objectlist = [(o.category(), o.xywh()) for frames in framelist for im in frames for o in im.objects()]

    Notes:
      - Some videos with a native framerate different from 30Hz may generate non-pixel accurate frames due to rounding errors in the ffmpeg fps filter chain
      - The videos are loaded with a framerate of 30Hz.  Boxes are annotated at 1Hz.  Objects are linearly interpolated to export at 30Hz
        but this means that some frames that have interpolated boxes may no longer precisely fall on the object
      - To recover the precise frame annotated once per second, use v.frame(n).  For example, for the frame at timestamp 2000ms, use v.frame(int(30*(2000/1000)))
      - Try v.show(timestamp=True) to show the frame index overlay along with the boxes to see which frame is being displayed
      - Delay the framerate conversion as late in the filter chain as possible (e.g. after trackclip)

    """
    def __init__(self, datadir=None, redownload=False):
        datadir = tocache('youtubeBB') if datadir is None else datadir
        
        # Download        
        self._datadir = remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            if not isinstalled('gunzip'):
                raise ValueError('Downloading requires the gunzip utility on the command line.  On Ubuntu: "sudo apt install gunzip"')                
            
            for url in URLS:
                vipy.downloader.download(url, os.path.join(self._datadir, filetail(url)))
                os.system('gunzip %s' % (os.path.join(self._datadir, filetail(url))))
            open(os.path.join(self._datadir, '.complete'), 'a').close()
            
        csv = readcsv(os.path.join(self._datadir, 'yt_bb_detection_train.csv'))

        # https://research.google.com/youtube-bb/download.html
        #youtube_id - same as above.
        #timestamp_ms - same as above.
        #class_id - same as above.
        #class_name - same as above.
        #object_id - (integer) an identifier of the object in the video. (see note below)
        #object_presence - same as above.
        #xmin - (float) a [0.0, 1.0] number indicating the left-most location of the bounding box in coordinates relative to the frame size.
        #xmax - (float) a [0.0, 1.0] number indicating the right-most location of the bounding box in coordinates relative to the frame size.
        #ymin - (float) a [0.0, 1.0] number indicating the top-most location of the bounding box in coordinates relative to the frame size.
        #ymax - (float) a [0.0, 1.0] number indicating the bottom-most location of the bounding box in coordinates relative to the frame size.

        # Notes:
        # - (xmin, ymin, xmax, ymax) = (-1,-1,-1,-1) if object_presence == 'absent'
        # - All framerates are defined relative to 30Hz videos
        # - Keyframes are sampled once every second.  Linear interpolation of boxes may be noisy.  
        
        youtubeids = list(set([x[0] for x in csv]))
        d_youtubeid_to_objectids = {k:set(x[4] for x in v) for (k,v) in groupbyasdict(csv, lambda x: x[0]).items()}
        d_youtubeid_objectid_to_bboxes = {k:[(float(x[1]), x[3], (float(x[6]),float(x[8]),float(x[7]),float(x[9]))) for x in v] for (k,v) in groupbyasdict(csv, lambda x: (x[0], x[4])).items()}  # (timestamp_ms, class_name, ulbr)
        
        loader = (lambda ytid, d_youtubeid_to_objectids=d_youtubeid_to_objectids, d_youtubeid_objectid_to_bboxes=d_youtubeid_objectid_to_bboxes:
                  Scene(url='http://youtu.be/%s' % ytid, framerate=30.0,
                        tracks=[Track(category=d_youtubeid_objectid_to_bboxes[(ytid,o)][0][1],
                                      keyframes=[int(float(ts)*(30/1000)) for (ts, c, ulbr) in d_youtubeid_objectid_to_bboxes[(ytid,o)] if ulbr[0]>=0],
                                      boxes=[Detection(category=c, ulbr=ulbr, normalized_coordinates=True) for (ts, c, ulbr) in d_youtubeid_objectid_to_bboxes[(ytid,o)] if ulbr[0]>=0])
                                for o in d_youtubeid_to_objectids[ytid]]))
                  
        super().__init__(youtubeids, id='youtubeBB', loader=loader)


        



import os
import vipy
from vipy.util import remkdir, filetail, filepath, tocache


URL = 'https://www.robots.ox.ac.uk/~vgg/data/dtd/download/dtd-r1.0.1.tar.gz'


class D2D(vipy.dataset.Dataset):
    """Describable Textures Dataset: https://www.robots.ox.ac.uk/~vgg/data/dtd/"""
    def __init__(self, datadir=None, redownload=False):
        # Download (if not cached)
        datadir = tocache('d2d') if datadir is None else datadir
        
        self._datadir = remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(URL, self._datadir, sha1=None)            

        # Create dataset
        imlist = tuple((f, filetail(filepath(f))) for (k,f) in enumerate(sorted(vipy.util.findimages(self._datadir))))
        loader = lambda x: vipy.image.ImageCategory(filename=x[0], category=x[1])
        super().__init__(imlist, id='d2d', loader=loader)

        open(os.path.join(self._datadir, '.complete'), 'a').close()



import os
import vipy
import torchvision.datasets


class CelebA(vipy.dataset.Dataset):
    """A thin wrapper around torchvision.datasets to import into vipy.dataset format.  
    
    https://docs.pytorch.org/vision/0.21/generated/torchvision.datasets.CelebA.html#torchvision.datasets.CelebA

    Requires gdown
    """
    
    def __init__(self, rootdir, download=False, split='train'):
        dataset = torchvision.datasets.CelebA(rootdir, download=download or not os.path.exists(os.path.join(rootdir, 'celeba/img_align_celeba.zip')), split=split, target_type=['identity','bbox'])

        # Boxes are wrong: https://github.com/pytorch/vision/issues/9008 
        #loader = lambda r: (vipy.image.Scene(objects=[vipy.object.Detection(category=str(int(r[1][0])), xywh=list(r[1][1]))]) if r[1] is not None else vipy.image.Image()).loader(vipy.image.Image.PIL_loader, r[0])

        # Fallback on category only (there is no mapping from index string to celebrity name)
        loader = lambda r: (vipy.image.ImageCategory(category=str(int(r[1][0]))) if r[1] is not None else vipy.image.Image()).loader(vipy.image.Image.PIL_loader, r[0])     
        super().__init__(dataset, id='celebA:'+split, loader=loader)

        



import os
import vipy


TRAIN_IMG_2021_URL = 'https://ml-inat-competition-datasets.s3.amazonaws.com/2021/train.tar.gz'
TRAIN_IMG_2021_MD5 = 'e0526d53c7f7b2e3167b2b43bb2690ed'
TRAIN_ANNO_2021_URL = 'https://ml-inat-competition-datasets.s3.amazonaws.com/2021/train.json.tar.gz'
TRAIN_ANNO_2021_MD5 = '38a7bb733f7a09214d44293460ec0021'
VAL_IMG_2021_URL = 'https://ml-inat-competition-datasets.s3.amazonaws.com/2021/val.tar.gz'
VAL_IMG_2021_MD5 = 'f6f6e0e242e3d4c9569ba56400938afc'
VAL_ANNO_2021_URL = 'https://ml-inat-competition-datasets.s3.amazonaws.com/2021/val.json.tar.gz'
VAL_ANNO_2021_MD5 = '4d761e0f6a86cc63e8f7afc91f6a8f0b'


class iNaturalist2021(vipy.dataset.Dataset):
    """Project: https://github.com/visipedia/inat_comp/tree/master/2021"""
    def __init__(self, datadir=None, imageurl=TRAIN_IMG_2021_URL, imagemd5=TRAIN_IMG_2021_MD5, annourl=TRAIN_ANNO_2021_URL, annomd5=TRAIN_ANNO_2021_MD5, name='inaturalist', redownload=False):

        datadir = vipy.util.tocache('iNaturalist2021') if datadir is None else datadir
        
        self._datadir = vipy.util.remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(imageurl, self._datadir, md5=imagemd5)
            vipy.downloader.download_and_unpack(annourl, self._datadir, md5=annomd5)

        json = vipy.util.readjson(os.path.join(self._datadir, '2021', vipy.util.filetail(annourl)[:-7]))  # remove trailing file extension (val.json.tar.gz -> val.json)

        d_imageid_to_filename = {x['id']:os.path.join(self._datadir, '2021', x['file_name']) for x in json['images']}
        d_imageid_to_annotation = {iid:a[0] for (iid,a) in vipy.util.groupbyasdict(json['annotations'], lambda x: x['image_id']).items()}  # one annotation per image
        d_categoryid_to_category = {x['id']:x['name'] for x in json['categories']}
        
        imlist = [(f,iid) for (iid,f) in d_imageid_to_filename.items()]

        loader = lambda x,d_imageid_to_annotation=d_imageid_to_annotation, d_categoryid_to_category=d_categoryid_to_category: vipy.image.ImageCategory(filename=x[0],
                                                    category=d_categoryid_to_category[d_imageid_to_annotation[x[1]]['category_id']] if x[1] in d_imageid_to_annotation else None,
                                                    attributes={'category_id': d_imageid_to_annotation[x[1]]['category_id']} if x[1] in d_imageid_to_annotation else None)
                                                    
        super().__init__(imlist, id=name, loader=loader)

        open(os.path.join(self._datadir, '.complete'), 'a').close()
        
    def trainset(self):
        return self.id('inaturalist:train')

    def valset(self):
        return iNaturalist2021(self._datadir, VAL_IMG_2021_URL, VAL_IMG_2021_MD5, VAL_ANNO_2021_URL, VAL_ANNO_2021_MD5, name='inaturalist:val')

    



import os
import vipy
from vipy.util import readcsv, tocache
from vipy.image import Scene
from vipy.object import Detection

URL = 'http://vis-www.cs.umass.edu/fddb/originalPics.tar.gz'
FOLDS_URL = 'http://vis-www.cs.umass.edu/fddb/FDDB-folds.tgz'
FOLDS_SHA1 = '94ce19ba3348425dfbb8bcbe55802490f8f152f8'


class FDDB(object):
    """Manages the FDDB dataset: http://vis-www.cs.umass.edu/fddb"""

    def __init__(self, datadir=None, redownload=False):
        datadir = tocache('fddb') if datadir is None else datadir
        
        self._datadir = vipy.util.remkdir(datadir)
        self._folds_dir = vipy.util.remkdir(os.path.join(self._datadir, 'FDDB-folds'))

        if redownload or os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(URL, self._datadir)                        
            #raise ValueError('Download FDDB dataset manually, with "wget %s -O %s; cd %s; tar zxvf %s"' % (URL, os.path.join(self._datadir, 'originalPics.tar.gz'), self._datadir, 'originalPics.tar.gz'))

            vipy.downloader.download_and_unpack(FOLDS_URL, self._folds_dir)
            #raise ValueError('Download FDDB-folds dataset manually, with "wget %s -O %s; cd %s; tar zxvf %s"' % (FOLDS_URL, os.path.join(self._datadir, 'FDDB-folds.tgz'), self._datadir, 'FDDB-folds.tgz'))

        open(os.path.join(self._datadir, '.complete'), 'a').close()
            
    def __repr__(self):
        return str("<vipy.dataset.fddb: '%s'>" % self._datadir)

    def fold(self, foldnum=1):
        """Return the foldnum as a list of vipy.image.Scene objects, each containing all vipy.object.Detection faces in the current image"""
        # fold_file = os.path.join(self._folds_dir, 'FDDB-fold-%02d.txt' % foldnum)
        k = 0
        rows = readcsv(os.path.join(self._folds_dir, 'FDDB-fold-%02d-ellipseList.txt' % foldnum), separator=' ')
        imscenes = []
        while k < len(rows):
            filename = rows[k][0]
            num_faces = int(rows[k + 1][0])
            bbox = [rows[j] for j in range(k + 2, k + 2 + num_faces)]
            k = k + 2 + len(bbox)

            # This ignores the rotation
            ims = Scene(filename=os.path.join(self._datadir, '%s.jpg' % filename), objects=[Detection('face', xcentroid=bb[3], ycentroid=bb[4], width=2 * float(bb[1]), height=2 * float(bb[0])) for bb in bbox])
            imscenes.append(ims)
        return imscenes



import os
import vipy


# Request from dataset organizers
URL = 'https://s3plus.meituan.net/v1/${KEY}/foodai-workshop-challenge/Food2k_complete.tar.gz'
INDEX = 'https://s3plus.meituan.net/v1/${KEY}/foodai-workshop-challenge/Food2k_label.tar.gz'
LABELS = 'http://123.57.42.89/Large-Scale_Food_Recognition_via_Deep_Progressive_Self-Transformer_Network/Supplementary%20tables.pdf'

class Food2k(vipy.dataset.Dataset):
    """Project: http://123.57.42.89/FoodProject.html"""
    def __init__(self, datadir=None, redownload=False):
        datadir = tocache('food2k') if datadir is None else datadir        

        # Download
        self._datadir = vipy.util.remkdir(datadir)        
        if redownload or not os.path.exists(os.path.join(self._datadir, '.complete')):
            vipy.downloader.download_and_unpack(URL, self._datadir)
            vipy.downloader.download_and_unpack(INDEX, self._datadir)

        d_idx_to_category = vipy.util.readjson(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'food2k.json'))  # scraped from PDF
        d_filename_to_idx = {vipy.util.filetail(f):str(v) for (f,v) in vipy.util.readcsv(os.path.join(self._datadir, 'Food2k_label', 'train_finetune.txt'), separator=' ')}
        loader = lambda f: vipy.image.ImageCategory(filename=f, category=d_idx_to_category[d_filename_to_idx[vipy.util.filetail(f)]])
        imlist = [f for f in vipy.util.findimages(os.path.join(datadir)) if vipy.util.filetail(f) in d_filename_to_idx]
        super().__init__(imlist, id='food2k', loader=loader)

        open(os.path.join(self._datadir, '.complete'), 'a').close()








import os
import vipy


class CAP_classification_clip(vipy.dataset.Dataset):
    URL = 'https://consented-activities-of-people.s3.us-west-2.amazonaws.com/train/cap_classification_clip.tar.gz'
    MD5 = '54315e2ce204f0dbbe298490a63b5b3b'
    def __init__(self, datadir):
        if not os.path.exists(os.path.join(datadir, 'cap_classification_clip', 'README.md')):
            vipy.downloader.download_and_unpack(CAP_classification_clip.URL, datadir)

        annotations = os.path.join(datadir, 'cap_classification_clip', 'annotations')
        super().__init__(vipy.dataset.Dataset.from_directory(annotations, filetype='json'), id='cap')


class CAP_classification_pad(vipy.dataset.Dataset):
    URL = 'https://consented-activities-of-people.s3.us-west-2.amazonaws.com/train/cap_classification_pad.tar.gz'
    MD5 = 'fbdc75e6ef10b874ddda20ee9765a710'
    def __init__(self, datadir):
        if not os.path.exists(os.path.join(datadir, 'cap_classification_pad', 'README.md')):
            vipy.downloader.download_and_unpack(CAP_classification_pad.URL, datadir)

        annotations = os.path.join(datadir, 'cap_classification_pad', 'annotations')
        super().__init__(vipy.dataset.Dataset.from_directory(annotations, filetype='json'), id='cap')

        
class CAP_detection(vipy.dataset.Dataset):
    URL = 'https://dl.dropboxusercontent.com/s/db11zv0gcycu384/cap_detection_handheld_val.tar.gz.72f58e69582c17dd366d3c7e85cf0da8'
    MD5 = '72f58e69582c17dd366d3c7e85cf0da8'
    def __init__(self, datadir):
        if not os.path.exists(os.path.join(datadir, 'cap_detection_handheld_val', 'README.md')):
            vipy.downloader.download(CAP_detection.URL, os.path.join(datadir, 'cap_detection_handheld_val.tar.gz'), md5=CAP_detection.MD5)
            vipy.downloader.unpack(os.path.join(datadir, 'cap_detection_handheld_val.tar.gz'), datadir)

        annotations = os.path.join(datadir, 'cap_detection_handheld_val', 'annotations')
        super().__init__(vipy.dataset.Dataset.from_directory(annotations, filetype='json'), id='cap_detection:val')
    



import vipy
import concurrent.futures as cf
import collections
import itertools
import threading
from queue import Queue


def map(f, ingen, reducer=None):
    """Apply the function f to each element in the iterator, returning the results unordered.  Function cannot be anonymous if cf executor is multiprocess """
    assert vipy.globals.cf() is not None, "vipy.globals.cf() executor required - Try 'with vipy.globals.parallel(n=4): vipy.parallel.map(...)' "    
    assert callable(f)    
    results = vipy.globals.cf().map(f, ingen)  # not order preserving
    return reducer(results) if reducer else results

def identity(x):
    return x


def iter(ingen, mapper=identity, bufsize=1024, progress=False, accepter=None):
    """Return an iterator on the input generator that will apply the mapper to each object and yield only those elements where the accepter is true

    ```python
    with vipy.globals.parallel(8):
        for im in vipy.parallel.iter(vipy.dataset.registry('yfcc100m_url:train'), mapper=lambda im: im.try_download(), accepter=lambda im: im.is_downloaded()):
            print(im)
    ```

    Most common use cases will use `vipy.dataset.Dataset.minibatch` or `vipy.dataset.Dataset.__parallel_iter__` instead of using this iterator directly

    Args:
        ingen [generator]:
        mapper [callable]: A function applied to each element before yielding
        bufsize [int]: The maximum size of the parallel queue used by producers.  
        accepter [callable]: A function which returns true or false, such that the iterator only yields elements for which the accepter returns true

    Returns:
        An iterator that yields mapped and accepted elements from ingen, whre mapping is performed in parallel by vipy.parallel.cf() executor

    """
    
    assert vipy.globals.cf(), "vipy.globals.cf() executor required - Try 'with vipy.globals.parallel(n=4): result = [x for x in vipy.parallel.iter(...)]' "    
    assert callable(mapper)
    assert accepter is None or callable(accepter)
    assert bufsize>0
    
    e = vipy.globals.cf()
    q = Queue()
    sem = threading.BoundedSemaphore(bufsize)
    
    if progress:
        vipy.util.try_import('tqdm','tqdm'); from tqdm import tqdm;
        ingen = tqdm(ingen, total=len(ingen) if hasattr(ingen, '__len__') else None)
    
    # Producer worker: this is useful for filling the pipeline while waiting on GPU I/O    
    def _producer():
        for i in ingen:
            sem.acquire()  # block when buffer is full
            f = e.submit(mapper, i)
            def _callback(fut, sem=sem, q=q):
                q.put(fut.result())      
                sem.release()
            f.add_done_callback(_callback)
        for k in range(bufsize):
            sem.acquire()  # wait until all callbacks have fired
        q.put(None)
        
    threading.Thread(target=_producer, daemon=True).start()

    # Consumer loop: yield loaded elements from the producers
    while True:        
        res = q.get()
        if res is None:
            break
        if accepter is None or accepter(res):
            yield res
    

def ordered_map(f, ingen):
    """Apply the function f to each element in the iterator, returning tuples (order, result) results sorted by order"""
    assert vipy.globals.cf(), "vipy.globals.cf() executor required - Try 'with vipy.globals.parallel(n=4): vipy.parallel.ordered_map(...)' "    
    assert callable(f), "mapper required"    
    return sorted(vipy.globals.cf().map(f, enumerate(ingen)), key=lambda x: x[0])  # expensive



    
    
